import math
import random
import time
from typing import Dict, List

import torch
from torch import Tensor
from tqdm import tqdm


def softmax(x: List[float]) -> List[float]:
    """
    Computes the softmax of a list of floats.

    Args:
        x: A list of floats.

    Returns:
        A list of floats representing the softmax of the input list.
    """
    if len(x) == 0:

        return []
    # 计算最大值，避免溢出
    max_val = max(x)
    # 计算指数和
    sum_exp = sum(math.exp(i - max_val) for i in x)
    # 计算softmax
    return [math.exp(i - max_val) / sum_exp for i in x]


def compress_logits(logits: torch.Tensor, top_k: int = 50, top_p: float = 0.99999) -> List[Dict[str, List[float]]]:
    """
    压缩logits矩阵，保留top_k和top_p概率的logits。

    Args:
        logits (torch.Tensor): 输入的logits矩阵，形状为(seq_len, vocab_size)。
        top_k (int): 保留的最大top_k个logits。
        top_p (float): 保留的累计概率阈值。

    Returns:
        List[Dict[str, List[float]]]: 压缩后的logits列表，每个元素为一个字典，
                                       包含保留的logits及其索引。
    """
    # 对logits进行softmax操作计算概率
    probs = torch.nn.functional.softmax(logits, dim=-1)

    # 获取top_k的概率和索引
    top_k_values, top_k_indices = torch.topk(probs, top_k, dim=-1)

    # 对top_k值进行排序
    sorted_values, sorted_indices = torch.sort(top_k_values, descending=True, dim=-1)

    # 计算累积概率
    cumulative_probs = torch.cumsum(sorted_values, dim=-1)
    mask = cumulative_probs <= top_p
    # 确保每个序列中mask中的True个数多一个
    true_counts = mask.sum(dim=-1, keepdim=True)
    next_indices = torch.clamp(true_counts, max=top_k - 1)
    mask.scatter_(1, next_indices, True)

    # 生成压缩后的logits列表
    compressed_logits = []
    for i in range(logits.size(0)):
        valid_indices = sorted_indices[i][mask[i]]
        compressed_logits.append({
            'indices': top_k_indices[i][valid_indices].tolist(),
            'values': logits[i][top_k_indices[i][valid_indices]].tolist()  # 保存原始logits值
        })

    return compressed_logits


def compress_probs(logits: torch.Tensor, top_k: int = 50, vocab_size=32000) -> List[Dict[str, List[float]]]:
    """
    压缩logits矩阵，保留top_k概率并保存剩余概率之和。

    Args:
        logits (torch.Tensor): 输入的logits矩阵，形状为(seq_len, vocab_size)。
        top_k (int): 保留的最大top_k个概率。

    Returns:
        List[Dict[str, List[float]]]: 压缩后的logits列表，每个元素为一个字典，
                                       包含保留的概率及其索引和剩余概率之和。
    """
    # 对logits进行softmax操作计算概率
    probs = torch.nn.functional.softmax(logits, dim=-1)

    # 获取top_k的概率和索引
    top_k_values, top_k_indices = torch.topk(probs, top_k, dim=-1)

    # 计算剩余概率之和
    remaining_probs_sum = 1 - top_k_values.sum(dim=-1)

    # 生成压缩后的概率列表
    compressed_probs = []
    for i in range(logits.size(0)):
        compressed_probs.append({
            'indices': top_k_indices[i].tolist(),
            'values': top_k_values[i].tolist(),  # 保存概率值
            'remaining_probs_sum': remaining_probs_sum[i].item()  # 剩余概率之和
        })
    return compressed_probs


def recover_margin_cd(
    compressed_probs: List[List[Dict]],
    input_probs: Tensor,
    margin_logp_every: List[List[Dict]],
    threshold_p: float = 0.1,
    do_sample=True,
    mini_vocab_size: int = 50,
    use_softmax=True,
):
    margin_cd = []
    batch_size, seq_len, vocab_size = input_probs.shape
    for batch_idx in range(batch_size):
        margin_cd_batch = []
        seq_len = len(compressed_probs[batch_idx])
        for seq_len_idx in range(seq_len):
            # 1. 索引
            probs_position = compressed_probs[batch_idx][seq_len_idx]
            margin_logp_position = margin_logp_every[batch_idx][seq_len_idx]

            # 2. 先找概率大于 threshold_p 的所有token （v_head）
            v_head_tokens = [vocab_idx
                             for (vocab_idx, p) in zip(probs_position['indices'], probs_position['values'])
                             if p >= threshold_p]

            if len(v_head_tokens) == 0:
                v_head_tokens = [vocab_idx for (vocab_idx, p) in zip(probs_position['indices'], probs_position['values'])
                                 if p >= threshold_p * 0.1]

                if len(v_head_tokens) == 0:
                    v_head_tokens = [vocab_idx for (vocab_idx, p) in zip(probs_position['indices'], probs_position['values'])
                                     if p >= threshold_p * 0.1 * 0.1]

            if len(v_head_tokens) > mini_vocab_size:
                v_head_tokens = [
                    vocab_idx for (vocab_idx, p) in
                    sorted(list(zip(probs_position['indices'], probs_position['values'])),
                           key=lambda x: x[1], reverse=True)]

            # 3. 找出v_head 所对应的 margin_logp
            v_head_idxes = [i for i, w in enumerate(margin_logp_position['indices']) if w in v_head_tokens]
            v_head_mlogp = [mlogp for i, mlogp in enumerate(margin_logp_position['values']) if i in v_head_idxes]

            # 4. 根据margin_logp的大小进行排序
            v_head_token_mlogp = sorted(list(zip([margin_logp_position['indices'][j] for j in v_head_idxes], v_head_mlogp)), key=lambda x: x[1], reverse=True)
            v_head_idxes = [vhtm[0] for vhtm in v_head_token_mlogp]
            if use_softmax:
                v_head_mlogp = softmax([vhtm[1] for vhtm in v_head_token_mlogp])
            else:
                v_head_mlogp = [vhtm[1] for vhtm in v_head_token_mlogp]

            # 5. 写入结果
            # 5.1 如果不需要采样，则直接把分数最大的token当作概率为1的token
            if not do_sample:
                margin_cd_seq = {
                    "indices": v_head_token_mlogp[0][0],
                    "values": 1.0,
                    "remaining_probs_sum": 0.0}

            # 5.2 如果需要采样，且len(v_head_idxes)太小，则需要填充至指定的`mini_vocab_size`
            elif len(v_head_idxes) <= mini_vocab_size:
                pad_token_idx = random.sample(
                    list(set(range(vocab_size)) - set(v_head_tokens)),
                    mini_vocab_size - len(v_head_idxes))
                margin_cd_seq = {
                    "indices": [v[0] for v in v_head_token_mlogp] + pad_token_idx,
                    "values": v_head_mlogp + [0.0] * len(pad_token_idx),
                    "remaining_probs_sum": 0.0}
            else:
                # 不会有这种情况
                margin_cd_seq = {
                    "indices": v_head_token_mlogp[0][0],
                    "values": 1.0,
                    "remaining_probs_sum": 0.0}

            margin_cd_batch.append(margin_cd_seq)
        margin_cd.append(margin_cd_batch)
    return margin_cd


def load_input_and_target_probs_fast(
    compressed_probs: List[List[Dict]],
    input_probs: Tensor,
    labels: Tensor,     # [batch_size, seq_len]
    from_chosen_or_rejected=None,
    prob_stabilize: bool = True,
    atkd=False
):
    # 初始化数值
    device = labels.device
    batch_size, seq_len, vocab_size = input_probs.shape
    input_probs_ = torch.cat([input_probs, torch.zeros(batch_size, seq_len, 1).to(device)], dim=-1)
    min_vocab_size = len(compressed_probs[0][0]['indices'])
    target_probs = torch.zeros([batch_size, seq_len, min_vocab_size + 1]).to(device)
    output_probs = torch.zeros([batch_size, seq_len, min_vocab_size + 1]).to(device)
    if from_chosen_or_rejected is None:
        pass
    else:
        compressed_probs_new_all = compressed_probs.copy()
        for i in range(len(compressed_probs)):
            for j in range(len(compressed_probs[i])):
                compressed_probs_ = compressed_probs[i][j]
                label_start = torch.where(labels[i] != -100)[0][0].item()
                label = labels[i][label_start + j].item()
                if from_chosen_or_rejected == "chosen":
                    if compressed_probs_['indices'][0] == label:
                        pass
                    else:
                        chosen_vocab_index = label
                        original_chosen_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index == chosen_vocab_index]
                        rank1_p = compressed_probs_['values'][0]
                        if len(original_chosen_index) == 1:
                            original_chosen_index = original_chosen_index[0]
                            decrease_ratio = (1 - rank1_p) / (1 - compressed_probs_['values'][original_chosen_index])
                        else:
                            decrease_ratio = 1 - rank1_p
                        out_chosen_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index != chosen_vocab_index]

                        # 如果chosen token不在此表中，将概率最小的单词移除，并加入`remaining_probs_sum`中
                        if len(out_chosen_index) >= min_vocab_size:
                            out_chosen_index = out_chosen_index[: min_vocab_size - 1]
                            compressed_probs_["remaining_probs_sum"] = compressed_probs_["remaining_probs_sum"] + sum(compressed_probs_['values'][min_vocab_size - 1: ])
                        compressed_probs_new = {
                            "indices": [chosen_vocab_index] + [compressed_probs_['indices'][i] for i in out_chosen_index],
                            "values": [rank1_p] + [compressed_probs_['values'][i] * decrease_ratio for i in out_chosen_index],
                            "remaining_probs_sum": compressed_probs_["remaining_probs_sum"] * decrease_ratio}
                        compressed_probs_new_all[i][j] = compressed_probs_new

                elif from_chosen_or_rejected == "rejected":
                    if label not in compressed_probs_['indices']:
                        pass
                    else:
                        rejected_vocab_index = label
                        original_rejected_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index == rejected_vocab_index][0]
                        out_rejected_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index != rejected_vocab_index]
                        increase_ratio = 1 / max(0.01, 1 - compressed_probs_['values'][original_rejected_index])
                        compressed_probs_new = {
                            "indices": [rejected_vocab_index] + [compressed_probs_['indices'][i] for i in out_rejected_index],
                            "values": [0] + [compressed_probs_['values'][i] * increase_ratio for i in out_rejected_index],
                            "remaining_probs_sum": compressed_probs_["remaining_probs_sum"] * increase_ratio}
                        compressed_probs_new_all[i][j] = compressed_probs_new

    easy_xy = []
    if atkd:
        compressed_probs_new_all = compressed_probs.copy()
        for i in range(len(compressed_probs)):
            unc_list = []
            for j in range(len(compressed_probs[i])):
                # 获取 label token
                compressed_probs_ = compressed_probs[i][j]
                label_start = torch.where(labels[i] != -100)[0][0].item()
                if len(labels[i]) <= label_start + j:
                    continue
                label = labels[i][label_start + j].item()

                # 计算UNC
                label_index = [(i, idx) for (i, idx) in enumerate(compressed_probs_['indices']) if idx == label]
                if len(label_index) == 0:
                    unc_list.append(1)
                else:
                    label_index = label_index[0][0]
                    unc_list.append(1 - compressed_probs_['values'][label_index])

            # 排序
            unc_list = sorted(enumerate(unc_list), key=lambda x: x[1])
            easy_set = [k for (k, _) in unc_list[: len(unc_list) // 2]]
            easy_xy += [(i, k) for k in easy_set]

            # 重建概率分布
            for j in range(len(compressed_probs[i])):
                if j not in easy_set:
                    continue
                else:
                    # 把 label token 的概率加入到 remaining_probs_sum中
                    compressed_probs_ = compressed_probs[i][j]
                    label_start = torch.where(labels[i] != -100)[0][0].item()
                    if len(labels[i]) <= label_start + j:
                        continue
                    label = labels[i][label_start + j].item()
                    label_index = [(i, idx) for (i, idx) in enumerate(compressed_probs_['indices']) if idx == label]
                    if len(label_index) == 0:
                        continue
                    else:
                        label_index = label_index[0][0]
                        rand_token_id = random.choice(list(set(range(vocab_size)) - set(compressed_probs_['indices'])))
                        compressed_probs_new = {
                            "indices": [k for k in compressed_probs_['indices'] if k != label] + [rand_token_id],
                            "values": [v for (k, v) in enumerate(compressed_probs_['values']) if k != label_index] + [0],
                            "remaining_probs_sum": compressed_probs_["remaining_probs_sum"] + compressed_probs_["values"][label_index]
                        }
                        compressed_probs_new_all[i][j] = compressed_probs_new
        compressed_probs = compressed_probs_new_all

    for i in range(batch_size):
        # 1. 边界检查
        # 1.1 将压缩张量进行还原，初始化
        compressed_probs_values = torch.Tensor(
            [position_compressed['values']
            for position_compressed in compressed_probs[i]]).to(device)  # seq_len, mini_vocab_size
        if 'remaining_probs_sum' not in compressed_probs[i][0].keys() or None in [position_compressed["remaining_probs_sum"] for position_compressed in compressed_probs[i]]:
            compressed_remaining_probs = torch.zeros(len(compressed_probs[i])).to(device)
        else:
            compressed_remaining_probs = torch.Tensor(
                [position_compressed["remaining_probs_sum"]
                for position_compressed in compressed_probs[i]]).to(device)
        compressed_probs_indices = torch.Tensor(
            [position_compressed['indices'] for position_compressed in compressed_probs[i]]).to(device)  # seq_len, mini_vocab_size

        # 1.2 先找出压缩张量中超出vocab_size边界的位置
        out_vocab_position = compressed_probs_indices >= vocab_size

        # 1.3 提取超出vocab_size边界的位置，并将这些位置的变量加到`remaining_probs_sum`中
        masked_floats = torch.where(out_vocab_position, compressed_probs_values, torch.tensor(0.0))
        compressed_remaining_probs += masked_floats.sum(dim=1)

        # 1.4 将超出vocab_size边界的位置的indices和values设置为0
        compressed_probs_indices[out_vocab_position] = vocab_size
        compressed_probs_values[out_vocab_position] = 0

        # 2. 赋值
        l, r = torch.where(labels[i] != -100)[0].min().item(), torch.where(labels[i] != -100)[0].max().item()
        target_probs[i][l: r + 1][:len(compressed_probs_values), :min_vocab_size] += compressed_probs_values[: r + 1 - l]
        target_probs[i][l: r + 1][:len(compressed_remaining_probs), min_vocab_size] += compressed_remaining_probs[: r + 1 - l]
        output_probs[i][l: r + 1][:len(compressed_probs_values), :min_vocab_size] = torch.gather(input_probs_[i][labels[i] != -100], 1, compressed_probs_indices[: r + 1 - l].to(torch.int64))
        output_probs[i][l: r + 1][:len(compressed_remaining_probs), min_vocab_size] = 1 - output_probs[i][l: r + 1][:len(compressed_probs_values), :min_vocab_size].sum(-1)
        pass
    # 稳定数值
    if prob_stabilize:
        output_probs[output_probs < 1e-8] = 1e-8
        output_probs[output_probs > (1 - 1e-8)] = 1 - 1e-8
        target_probs[target_probs < 1e-8] = 1e-8
        target_probs[target_probs > (1 - 1e-8)] = 1 - 1e-8

    # easy mask (only for atkd)
    if atkd:
        x_coords, y_coords = zip(*easy_xy)
        x_coords = torch.tensor(x_coords)
        y_coords = torch.tensor(y_coords)
        is_easy_mask = torch.full_like(target_probs, fill_value=False, dtype=torch.bool)
        is_easy_mask[x_coords, y_coords] = True
        return output_probs, target_probs, is_easy_mask

    return output_probs, target_probs


def load_input_and_target_probs(
    compressed_probs: List[List[Dict]],
    input_probs: Tensor,
    labels: Tensor,
    from_chosen_or_rejected=None,
):

    # sum([(i.item() == j.item()) for (i, j) in zip(input_probs[0].argmax(-1), labels[0]) if j ！= -100]);
    # sum([(i == j) for (i, j) in zip([i['indices'][0] for i in compressed_probs[0]], [j for j in labels[0] if j != -100])]);
    # sum([(i == j) for (i, j) in zip([i['indices'][0] for i in compressed_probs[0]], [i for (i, j) in zip(input_probs[0].argmax(-1), labels[0]) if j != -100])])
    batch_size, seq_len, vocab_size = input_probs.shape
    min_vocab_size = len(compressed_probs[0][0]['indices'])
    target_probs = torch.zeros([batch_size, seq_len, min_vocab_size + 1])
    output_probs = torch.zeros([batch_size, seq_len, min_vocab_size + 1])
    for i in range(batch_size):
        k = 0
        for j in range(seq_len):
            if labels[i][j] == -100:
                continue
            if k >= len(compressed_probs[i]):
                # output_probs[i][j] = 1 / min_vocab_size
                # target_probs[i][j] = 1 / min_vocab_size
                pass
            else:
                compressed_probs_ = compressed_probs[i][k]

                # 将chosen设置为rank1
                if from_chosen_or_rejected == "chosen":
                    if compressed_probs_['indices'][0] == labels[i][j]:
                        pass
                    else:
                        chosen_vocab_index = labels[i][j]
                        original_chosen_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index == chosen_vocab_index]
                        rank1_p = compressed_probs_['values'][0]
                        if len(original_chosen_index) == 1:
                            original_chosen_index = original_chosen_index[0]
                            decrease_ratio = (1 - rank1_p) / (1 - compressed_probs_['values'][original_chosen_index])
                        else:
                            decrease_ratio = 1 - rank1_p
                        out_chosen_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index != chosen_vocab_index]
                        if len(out_chosen_index) >= min_vocab_size:
                            out_chosen_index = out_chosen_index[: min_vocab_size - 1]
                            compressed_probs_["remaining_probs_sum"] = compressed_probs_["remaining_probs_sum"] + sum(compressed_probs_['values'][min_vocab_size - 1: ])
                        compressed_probs_new = {
                            "indices": [chosen_vocab_index.item()],
                            "values": [rank1_p],
                            "remaining_probs_sum": compressed_probs_["remaining_probs_sum"] * decrease_ratio}
                        compressed_probs_new['indices'] = \
                            compressed_probs_new['indices'] + \
                            [compressed_probs_['indices'][i] for i in out_chosen_index]
                        compressed_probs_new['values'] = \
                            compressed_probs_new['values'] + \
                            [compressed_probs_['values'][i] * decrease_ratio for i in out_chosen_index]
                        compressed_probs_ = compressed_probs_new
                elif from_chosen_or_rejected == "rejected":
                    if labels[i][j] not in compressed_probs_['indices']:
                        pass
                    else:
                        rejected_vocab_index = labels[i][j]
                        original_rejected_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index == rejected_vocab_index][0]
                        out_rejected_index = [index for (index, vocab_index) in enumerate(compressed_probs_['indices']) if vocab_index != rejected_vocab_index]
                        increase_ratio = 1 / max(0.01, 1 - compressed_probs_['values'][original_rejected_index])
                        compressed_probs_new = {
                            "indices": [rejected_vocab_index.item()],
                            "values": [0],
                            "remaining_probs_sum": compressed_probs_["remaining_probs_sum"] * increase_ratio}
                        compressed_probs_new['indices'] = \
                            compressed_probs_new['indices'] + \
                            [compressed_probs_['indices'][i] for i in out_rejected_index]
                        compressed_probs_new['values'] =  \
                            compressed_probs_new['values'] + \
                            [compressed_probs_['values'][i] * increase_ratio for i in out_rejected_index]
                        compressed_probs_ = compressed_probs_new

                if max(compressed_probs_['indices']) < vocab_size:
                    output_probs[i][j][: min_vocab_size] = input_probs[i][j][compressed_probs_['indices']]
                    output_probs[i][j][min_vocab_size] = 1 - output_probs[i][j][: min_vocab_size].sum()
                    target_probs[i][j][: min_vocab_size] = torch.tensor(compressed_probs_['values'])
                    target_probs[i][j][min_vocab_size] = compressed_probs_['remaining_probs_sum']
                    k = k + 1
                else:
                    # 词表越界
                    out_vocab_index = [index for (index, word_idx) in enumerate(compressed_probs_['indices']) if word_idx >= vocab_size]
                    processed_compressed_probs = {
                        'indices': [word_idx for (index, word_idx)
                                    in enumerate(compressed_probs_['indices']) if index not in out_vocab_index],
                        'values': [value for (index, value)
                                   in enumerate(compressed_probs_['values']) if index not in out_vocab_index],
                        'remaining_probs_sum': compressed_probs_['remaining_probs_sum'] + \
                                               sum([value for (index, value)
                                                    in enumerate(compressed_probs_['values']) if index in out_vocab_index])
                    }
                    output_probs[i][j][: len(processed_compressed_probs['indices'])] = input_probs[i][j][processed_compressed_probs['indices']]
                    output_probs[i][j][min_vocab_size] = 1 - output_probs[i][j][: min_vocab_size].sum()
                    target_probs[i][j][: len(processed_compressed_probs['values'])] = torch.tensor(processed_compressed_probs['values'])
                    target_probs[i][j][min_vocab_size] = processed_compressed_probs['remaining_probs_sum']
                    k = k + 1
    output_probs[output_probs < 1e-7] = 1e-7
    output_probs[output_probs > (1 - 1e-7)] = 1 - 1e-7
    target_probs[target_probs < 1e-7] = 1e-7
    target_probs[target_probs > (1 - 1e-7)] = 1 - 1e-7
    return output_probs.to(input_probs.device), target_probs.to(input_probs.device)


def restore_logits(compressed_logits: List[List[Dict[str, List[float]]]], mask_labels, batch_size, seq_len: int, vocab_size: int) -> Tensor:
    """
    还原logits矩阵为原始形状。

    Args:
        compressed_logits (List[Dict[str, List[float]]]): 压缩后的logits列表。
        seq_len (int): 原始序列长度。
        vocab_size (int): 词表大小。

    Returns:
        torch.Tensor: 还原后的logits矩阵，形状为(seq_len, vocab_size)。
    """
    restored_logits = torch.full((batch_size, seq_len, vocab_size), - torch.inf)
    for batch_id, (batch, mask) in enumerate(zip(compressed_logits, mask_labels)):
        j = 0
        for i, mask_token in enumerate(mask):
            if mask_token == -100:
                restored_logits[batch_id, i, :] = 0
            elif j == len(batch):
                j += 1
                continue
            else:
                indices = batch[j]['indices']
                values = batch[j]['values']
                try:
                    restored_logits[batch_id, i, indices] = torch.tensor(values)
                except IndexError:
                    out_of_vocab_indexes = [i for (i, idx) in enumerate(indices) if idx >= vocab_size or idx < 0]
                    indices = [idx for (i, idx) in enumerate(indices) if i not in out_of_vocab_indexes]
                    values = [idx for (i, idx) in enumerate(values) if i not in out_of_vocab_indexes]
                    restored_logits[batch_id, i, indices] = torch.tensor(values)
                j += 1
    return restored_logits.to(torch.bfloat16)


if __name__ == '__main__':
    vocab_size = 100430
    seq_len = 768
    t = time.time()
    for i in tqdm(range(65600)):
        # 示例logits
        logits = torch.randn((768, vocab_size))  # (seq_len=5, vocab_size=10000)

        # 压缩logits
        compressed = compress_probs(logits, top_k=100)

        # 还原logits
        restored = restore_logits(compressed, seq_len=seq_len, vocab_size=vocab_size)
    print(time.time() - t)

