import copy
import json
import os.path
import time
import typing
import warnings
from dataclasses import dataclass, field
from functools import partial
from typing import Optional, List, Union

import numpy as np
import torch
import torch.distributed as dist
from accelerate import Accelerator
from accelerate import PartialState
from accelerate.utils import (
    is_npu_available,
    is_xpu_available, gather_object,
)
from datasets import load_dataset, load_from_disk, Dataset
from peft import LoraConfig
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from transformers import AutoTokenizer, HfArgumentParser, AutoModelForCausalLM, PreTrainedModel, \
    is_safetensors_available, BatchEncoding, DataCollatorForSeq2Seq
from transformers.utils import logging
from trl import AutoModelForCausalLMWithValueHead, AutoModelForSeq2SeqLMWithValueHead, PPOConfig, PPOTrainer, set_seed
from trl.core import logprobs_from_logits, stack_dicts, stats_to_np, WANDB_PADDING, convert_to_scalar, PPODecorators

if is_safetensors_available():
    pass
logger = logging.get_logger(__name__)
tqdm.pandas()


@dataclass
class ScriptArguments:
    use_seq2seq: bool = field(default=False, metadata={"help": "whether to use seq2seq"})
    trust_remote_code: bool = field(default=False, metadata={"help": "Enable `trust_remote_code`"})

    # LoraConfig
    use_peft: bool = field(default=False, metadata={"help": "whether to use peft"})
    lora_alpha: Optional[float] = field(default=16, metadata={"help": "the lora alpha parameter"})
    lora_r: Optional[int] = field(default=16, metadata={"help": "the lora r parameter"})

    # # PPO Config
    reward_level: Optional[str] = field(default="sentence")
    valid_dataset: Optional[str] = field(default="/data/spgao/datasets/datasets--argilla--dpo-mix-7k")
    test_dataset: Optional[str] = field(
        default="/data/spgao/alignment-handbook-cs/validate_by_rm/models/gpt4_1106_preview.json")
    pretraining_loss_weight: Optional[float] = field(default=1)

    # train and save
    num_train_epochs: Optional[int] = field(default=3)
    save_steps: Optional[int] = field(default=23)
    output_dir: Optional[str] = field(default="output")

    # distill setting
    max_new_tokens: Optional[int] = field(default=1024)
    dpo_teacher: Optional[str] = field(
        default="/data/spgao/alignment-handbook-cs/data/h2o-danube3-500m-base/dpkd_pa_nstart")
    ref_teacher: Optional[str] = field(
        default="/data/spgao/alignment-handbook-cs/data/h2o-danube3-500m-base/vanilla_kd_pa")


def batch_prompts(prompts, batch_size=16):
    return [prompts[i:i + batch_size] for i in range(0, len(prompts), batch_size)]


def build_dataset(query_dataset, dataset_num_proc):
    """
    Build dataset for training. This builds the dataset from `load_dataset`, one should
    customize this function to train the model on its own dataset.

    Args:
        query_dataset (`str`):
            The name of the dataset to be loaded.

    Returns:
        dataloader (`torch.utils.data.DataLoader`):
            The dataloader for the dataset.
    """
    # load imdb with datasets
    ds = load_dataset(query_dataset, split="train")

    def tokenize(sample):
        sample["query"] = tokenizer.apply_chat_template(sample['chosen'][: -1], add_generation_prompt=True,
                                                        tokenize=False)
        sample["response"] = sample['chosen'][-1]['content'] + tokenizer.eos_token
        sample["input_ids"] = tokenizer.encode(sample["query"])
        return sample

    ds = ds.map(tokenize, num_proc=dataset_num_proc)
    ds = ds.filter(lambda x: len(tokenizer.encode(x["query"])) < args.max_new_tokens and len(
        tokenizer.encode(x["response"])) < args.max_new_tokens)
    ds.set_format(type="torch")
    return ds


def collator(data):
    return {key: [d[key] for d in data] for key in data[0]}


def get_logps_for_query_response(model: PreTrainedModel, input_ids, labels):
    input_ids = input_ids.to(model.device)
    labels = labels.to(model.device)
    model.eval()
    with torch.no_grad():
        logits = model(input_ids).logits

    # shift
    logits = logits[:, : -1, :]
    input_ids = input_ids[:, 1:]
    labels = labels[:, 1:]

    # get logp
    log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
    gathered_log_probs = torch.gather(log_probs, -1, input_ids.unsqueeze(-1)).squeeze(-1)
    # masked_log_probs = torch.where(labels != -100, gathered_log_probs, torch.tensor(0.0))

    # get the position of the eos_token in response
    rst = []
    for (glp, label, eos_position) in zip(gathered_log_probs, labels,
                                          (labels == tokenizer.eos_token_id).int().argmax(-1)):
        position = (label != -100) & (label != tokenizer.pad_token_id)
        if eos_position != 0 and sum(position) < args.max_new_tokens:
            position[eos_position] = True
        if len(glp[position]) > args.max_new_tokens:
            print("cp")
        r = glp[position]
        if glp[position].dim() == 0:
            r = glp[position].unsqueeze(0)
        rst.append(r)

    return rst


class CustomPPOTrainer(PPOTrainer):
    def __init__(self, *kargs, **kwargs):
        """
        初始化 CustomPPOTrainer，保持与 PPOTrainer 的兼容性。
        """
        kwargs['dataset'] = kwargs.pop("query_dataset")
        sft_dataset: Dataset = kwargs.pop("sft_dataset")
        super().__init__(*kargs, **kwargs)
        self.is_deepspeed_enabled = getattr(self.accelerator.state, "deepspeed_plugin", None) is not None
        self.is_fsdp_enabled = getattr(self.accelerator.state, "fsdp_plugin", None) is not None
        self.writer = SummaryWriter(log_dir=args.output_dir)
        sft_dataset = sft_dataset.remove_columns(
            [c for c in sft_dataset.column_names if c not in ['query', 'response']])
        sft_dataset = sft_dataset.map(self.process_sft_row)
        self.sft_dataloader = torch.utils.data.DataLoader(
            sft_dataset,
            batch_size=self.config.batch_size,
            shuffle=True,
            drop_last=True,
            collate_fn=collator,
        )
        self.sft_collator = DataCollatorForSeq2Seq(tokenizer=tokenizer)

    def process_sft_row(self, row):
        query = self.tokenizer.encode(row['query'])
        response = self.tokenizer.encode(row['response'])
        input_ids = query + response
        labels = [-100] * len(query) + response
        return {"input_ids": input_ids, "labels": labels}

    def compute_rewards(
            self,
            scores: Union[torch.FloatTensor, List[torch.FloatTensor]],
            logprobs: torch.FloatTensor,
            ref_logprobs: torch.FloatTensor,
            masks: torch.LongTensor,
    ):
        """
        根据输入的 score 和 KL penalty 计算每个位置的奖励。
        支持两种 `score` 形式：
        1. [batch_size]: 将该奖励放在序列的最后一个时间步。
        2. [batch_size, seq_len]: 每个位置都有一个奖励值。如果 seq_len 不够，则使用 0 填充。

        Args:
            scores (torch.FloatTensor): 奖励的分数，可以是 [batch_size] 或 [batch_size, seq_len] 形式。
            logprobs (torch.FloatTensor): 模型的 log 概率，形状 [batch_size, seq_len]。
            ref_logprobs (torch.FloatTensor): 参考模型的 log 概率，形状 [batch_size, seq_len]。
            masks (torch.LongTensor): 用于掩码的张量，形状 [batch_size, seq_len]。

        Returns:
            torch.FloatTensor: 每个位置的总奖励，形状 [batch_size, seq_len]。
            torch.FloatTensor: 非 score 奖励 (KL penalty)，形状 [batch_size, seq_len]。
            torch.FloatTensor: KL penalty，形状 [batch_size, seq_len]。
        """
        rewards, non_score_rewards, kls = [], [], []

        for score, logprob, ref_logprob, mask in zip(scores, logprobs, ref_logprobs, masks):
            # 计算 KL 惩罚
            kl = self._kl_penalty(logprob, ref_logprob)
            kls.append(kl)

            # KL 惩罚作为负奖励
            non_score_reward = -self.kl_ctl.value * kl
            non_score_rewards.append(non_score_reward)
            reward = non_score_reward.clone()  # 复制 KL 惩罚作为初始奖励

            # 如果 score 是 [batch_size]，将奖励放在最后一个非掩码的位置
            if score.numel() == 1:
                last_non_masked_index = mask.nonzero()[-1]  # 找到最后一个非掩码的位置
                reward[last_non_masked_index] += score  # 在最后一个位置加上序列级别奖励
            else:
                # 如果 score 是 [batch_size, seq_len]，确保它与 non_score_reward 的形状相同
                # if score.size(0) < non_score_reward.size(0):
                #     # 使用 0 填充 score 使其长度与 non_score_reward 一致
                #     padding = non_score_reward.size(0) - score.size(0)
                #     score = torch.cat([score, torch.zeros(score.size(0), padding, device=score.device)], dim=1)
                try:
                    reward[mask != 0] = non_score_reward[mask != 0] + score  # 逐步奖励与 KL 惩罚相加
                except RuntimeError:
                    print("cp")

            rewards.append(reward)

        return torch.stack(rewards), torch.stack(non_score_rewards), torch.stack(kls)

    @PPODecorators.empty_device_cache()
    def train_minibatch(
            self,
            old_logprobs: torch.FloatTensor,
            values: torch.FloatTensor,
            logprobs: torch.FloatTensor,
            logits: torch.FloatTensor,
            vpreds: torch.FloatTensor,
            mask: torch.LongTensor,
            advantages: torch.FloatTensor,
            returns: torch.FloatTensor,
            pretraining_corpus: BatchEncoding = None,  # Added argument for pretraining labels
    ):
        """
        Train one PPO minibatch and incorporate the pretraining loss.

        Args:
            old_logprobs (`torch.FloatTensor`): Log probabilities of the old policy.
            values (`torch.FloatTensor`): Values of the value head.
            logprobs (`torch.FloatTensor`): Log probabilities of the current policy.
            logits (`torch.FloatTensor`): Logits from the model.
            vpreds (`torch.FloatTensor`): Predicted values.
            mask (`torch.LongTensor`): Attention masks.
            advantages (`torch.FloatTensor`): Advantage values.
            returns (`torch.FloatTensor`): Return values.
            pretraining_corpus (`BatchEncoding`): Ground truth labels for pretraining corpus.

        Returns:
            train_stats (dict[str, `torch.Tensor`]): Dictionary of training statistics.
        """
        self.model.train()

        # Step 1: Compute PPO Loss (same as before)
        loss_p, loss_v, train_stats = self.loss(
            old_logprobs, values, logits, vpreds, logprobs, mask, advantages, returns
        )

        # Step 2: Compute Pretraining Loss (cross-entropy loss on the pretraining labels)
        if pretraining_corpus is not None:
            pretraining_corpus = self.sft_collator(copy.deepcopy(pretraining_corpus)).to(self.current_device)
            _, pretraining_loss, _ = self.model(**pretraining_corpus)
        else:
            pretraining_loss = 0

        # Step 3: Combine PPO Loss and Pretraining Loss with a weighting factor
        pretraining_loss_weight = args.pretraining_loss_weight  # This should be added to the config
        total_loss = loss_p + loss_v + pretraining_loss_weight * pretraining_loss

        # Step 4: Backpropagation
        self.accelerator.backward(total_loss)
        if self.config.max_grad_norm is not None:
            if self.accelerator.sync_gradients:
                self.accelerator.clip_grad_norm_(self.model_params, self.config.max_grad_norm)

        # Step 5: Optimizer step
        self.optimizer.step()
        self.optimizer.zero_grad()

        return train_stats

    def step(
            self,
            queries: List[torch.LongTensor],
            responses: List[torch.LongTensor],
            scores: List[torch.FloatTensor], sft_batch: List[typing.Dict[str, str]] = None,
            response_masks: Optional[List[torch.LongTensor]] = None,
    ):
        """
        运行一个 PPO 优化步骤，支持 [batch_size] 或 [batch_size, seq_len] 形式的 score。
        处理不同长度的 scores，并使用填充处理变长序列。
        """
        bs = self.config.batch_size

        # 对输入的数据进行安全检查和处理
        queries, responses, scores, response_masks = self._step_safety_checker(
            bs, queries, responses, scores, response_masks
        )

        # 将 list of tensors 的 scores 填充到相同的长度
        if args.reward_level == "token":
            scores = [torch.tensor(score, device=self.current_device) for score in scores]
        else:
            scores = torch.tensor(scores, device=self.current_device)

        # 使用 score scaling 和 score clipping
        if self.config.use_score_scaling:
            # Score scaling
            if args.reward_level == "token":
                scores_mean, scores_std = self.running.update(torch.concat([s for s in scores if s.dim() != 0]))
            else:
                scores_mean, scores_std = self.running.update(scores)

            tensor_to_kwargs = dict(dtype=scores[0].dtype, device=scores[0].device)
            score_scaling_factor = self.running.std.clone().detach().to(**tensor_to_kwargs) + torch.finfo(scores[
                                                                                                              0].dtype).eps  # score_scaling_factor = torch.tensor(self.running.std).to(**tensor_to_kwargs) + torch.finfo(scores[0].dtype).eps
            if self.config.use_score_norm:
                if args.reward_level == "token":
                    scores = [(s - torch.tensor(self.running.mean).to(**tensor_to_kwargs)) / score_scaling_factor for s
                              in scores]
                else:
                    scores = (scores - torch.tensor(self.running.mean).to(**tensor_to_kwargs)) / score_scaling_factor
            else:
                if args.reward_level == "token":
                    scores = [s / score_scaling_factor for s in scores]
                else:
                    scores /= score_scaling_factor

        if self.config.score_clip is not None:
            # Score clipping
            scores_dtype = scores[0].dtype
            if args.reward_level == "token":
                scores = [torch.clip(s.float(), -self.config.score_clip, self.config.score_clip).to(dtype=scores_dtype)
                          for s in scores]
            else:
                scores = torch.clip(scores.float(), -self.config.score_clip, self.config.score_clip).to(
                    dtype=scores_dtype)
        timing = dict()
        t0 = time.time()

        t = time.time()

        model_inputs = self.prepare_model_inputs(queries, responses)

        if self.is_distributed:
            pad_first = self.tokenizer.padding_side == "left"

            model_inputs["input_ids"] = self.accelerator.pad_across_processes(
                model_inputs["input_ids"],
                dim=1,
                pad_index=self.tokenizer.pad_token_id,
                pad_first=pad_first,
            )
            model_inputs["attention_mask"] = self.accelerator.pad_across_processes(
                model_inputs["attention_mask"], dim=1, pad_index=0, pad_first=pad_first
            )
            if self.is_encoder_decoder:
                model_inputs["decoder_input_ids"] = self.accelerator.pad_across_processes(
                    model_inputs["decoder_input_ids"],
                    dim=1,
                    pad_index=self.tokenizer.pad_token_id,
                    pad_first=pad_first,
                )
                model_inputs["decoder_attention_mask"] = self.accelerator.pad_across_processes(
                    model_inputs["decoder_attention_mask"],
                    dim=1,
                    pad_index=0,
                    pad_first=pad_first,
                )

        model_inputs_names = list(model_inputs.keys())

        full_kl_penalty = self.config.kl_penalty == "full"

        with torch.no_grad():
            all_logprobs, logits_or_none, values, masks = self.batched_forward_pass(
                self.model,
                queries,
                responses,
                model_inputs,
                response_masks=response_masks,
                return_logits=full_kl_penalty,
            )
            with self.optional_peft_ctx():
                ref_logprobs, ref_logits_or_none, _, _ = self.batched_forward_pass(
                    self.model if self.is_peft_model else self.ref_model,
                    queries,
                    responses,
                    model_inputs,
                    return_logits=full_kl_penalty,
                )

        timing["time/ppo/forward_pass"] = time.time() - t

        with torch.no_grad():
            t = time.time()
            if full_kl_penalty:
                active_full_logprobs = logprobs_from_logits(logits_or_none, None, gather=False)
                ref_full_logprobs = logprobs_from_logits(ref_logits_or_none, None, gather=False)

                rewards, non_score_reward, kls = self.compute_rewards(
                    scores, active_full_logprobs, ref_full_logprobs, masks
                )
            else:
                rewards, non_score_reward, kls = self.compute_rewards(scores, all_logprobs, ref_logprobs, masks)
            timing["time/ppo/compute_rewards"] = time.time() - t

            t = time.time()
            values, advantages, returns = self.compute_advantages(values, rewards, masks)
            timing["time/ppo/compute_advantages"] = time.time() - t

        # upcast to float32 to avoid dataset issues
        batch_dict = {
            "queries": queries,
            "responses": responses,
            "logprobs": all_logprobs.to(torch.float32),
            "values": values.to(torch.float32),
            "masks": masks,
            "advantages": advantages,
            "returns": returns,
            "sft_corpus": sft_batch
        }
        batch_dict.update(model_inputs)

        t = time.time()
        all_stats = []
        early_stop = False
        for _ in range(self.config.ppo_epochs):
            if early_stop:
                break
            b_inds = np.random.permutation(bs)
            for backward_batch_start in range(0, bs, self.config.backward_batch_size):
                backward_batch_end = backward_batch_start + self.config.backward_batch_size
                backward_batch_inds = b_inds[backward_batch_start:backward_batch_end]

                for mini_batch_start in range(0, self.config.backward_batch_size, self.config.mini_batch_size):
                    mini_batch_end = mini_batch_start + self.config.mini_batch_size
                    mini_batch_inds = backward_batch_inds[mini_batch_start:mini_batch_end]
                    mini_batch_dict = {
                        "logprobs": batch_dict["logprobs"][mini_batch_inds],
                        "values": batch_dict["values"][mini_batch_inds],
                        "masks": batch_dict["masks"][mini_batch_inds],
                        # hacks: the queries and responses are ragged.
                        "queries": [batch_dict["queries"][i] for i in mini_batch_inds],
                        "responses": [batch_dict["responses"][i] for i in mini_batch_inds],
                        "advantages": batch_dict["advantages"][mini_batch_inds],
                        "returns": batch_dict["returns"][mini_batch_inds],
                        "sft_corpus": [batch_dict["sft_corpus"][i] for i in mini_batch_inds]
                    }
                    for k in model_inputs_names:
                        mini_batch_dict[k] = batch_dict[k][mini_batch_inds]
                    with self.accelerator.accumulate(self.model):
                        model_inputs = {k: mini_batch_dict[k] for k in model_inputs_names}

                        logprobs, logits, vpreds, _ = self.batched_forward_pass(
                            self.model,
                            mini_batch_dict["queries"],
                            mini_batch_dict["responses"],
                            model_inputs,
                            return_logits=True,
                        )
                        train_stats = self.train_minibatch(
                            mini_batch_dict["logprobs"],
                            mini_batch_dict["values"],
                            logprobs,
                            logits,
                            vpreds,
                            mini_batch_dict["masks"],
                            mini_batch_dict["advantages"],
                            mini_batch_dict["returns"],
                            mini_batch_dict["sft_corpus"]
                        )
                        all_stats.append(train_stats)

            # typically, early stopping is done at the epoch level
            if self.config.early_stopping:
                policykl = train_stats["policy/policykl"]
                early_stop = self._early_stop(policykl)
                if early_stop:
                    break

        timing["time/ppo/optimize_step"] = time.time() - t

        t = time.time()
        train_stats = stack_dicts(all_stats)

        # reshape advantages/ratios such that they are not averaged.
        train_stats["policy/advantages"] = torch.flatten(train_stats["policy/advantages"]).unsqueeze(0)
        train_stats["policy/advantages"] = torch.nan_to_num(train_stats["policy/advantages"], WANDB_PADDING)
        train_stats["policy/ratio"] = torch.flatten(train_stats["policy/ratio"]).unsqueeze(0)

        stats = self.record_step_stats(
            scores=torch.tensor([torch.sum(s) for s in scores]).to(self.current_device),
            logprobs=all_logprobs,
            ref_logprobs=ref_logprobs,
            non_score_reward=non_score_reward,
            train_stats=train_stats,
            kl_coef=self.kl_ctl.value,
            masks=masks,
            queries=queries,
            responses=responses,
            kls=kls,
        )
        # Gather/Reduce stats from all processes
        if self.is_distributed:
            stats = self.gather_stats(stats)
        stats = stats_to_np(stats)
        timing["time/ppo/calc_stats"] = time.time() - t
        stats["ppo/learning_rate"] = self.optimizer.param_groups[0]["lr"]

        # Update the KL control - multiply the batch_size by the number of processes
        self.kl_ctl.update(
            stats["objective/kl"],
            self.config.batch_size * self.accelerator.num_processes,
        )

        # Log the total ppo time
        timing["time/ppo/total"] = time.time() - t0
        stats.update(timing)

        # post-process stats for tensorboard and other loggers
        if self.config.log_with != "wandb":
            stats = convert_to_scalar(stats)

        if self.lr_scheduler is not None:
            self.lr_scheduler.step()

        return stats

    def save_model(self, output_dir, valid_prompts=None):
        if valid_prompts is not None:

            generation_kwargs = {
                "min_length": -1,
                "do_sample": False,
                "pad_token_id": tokenizer.eos_token_id,
                "max_new_tokens": args.max_new_tokens,
            }
            self.accelerator.wait_for_everyone()
            # Initialize the local response list
            local_prompts_responses = []

            # Split the valid_prompts to each process
            with self.accelerator.split_between_processes(valid_prompts) as v_prompts:
                responses = []
                prompt_batches = batch_prompts(v_prompts, batch_size=self.config.mini_batch_size * 2)
                for prompt_batch in tqdm(prompt_batches):
                    prompt_batch_tensor = [torch.LongTensor(self.tokenizer.encode(prompt)).to(self.current_device) for
                                           prompt in prompt_batch]
                    response_tensors = self.generate(
                        prompt_batch_tensor, return_prompt=False, generate_ref_response=False, **generation_kwargs
                    )
                    responses.extend(self.tokenizer.batch_decode(response_tensors))

                # Collect local prompts and responses
                local_prompts_responses.extend(
                    [{"prompt": prompt, "response": [[response]]} for (prompt, response) in zip(v_prompts, responses)]
                )

            # Create an empty list to hold gathered results from all processes
            gathered_prompts_responses = [None for _ in range(dist.get_world_size())]

            # Use dist.all_gather_object to gather responses from all processes
            dist.all_gather_object(gathered_prompts_responses, local_prompts_responses)

            if self.accelerator.is_main_process:
                if valid_prompts is not None:
                    valid_output = []
                    for gpr in gathered_prompts_responses:
                        valid_output.extend(gpr)
                    if not os.path.exists(output_dir):
                        os.mkdir(output_dir)
                    with open(os.path.join(output_dir, "valid_output.jsonl"), "w") as f:
                        f.writelines(
                            [json.dumps({"prompt": prompt, "response": [[response]]}, ensure_ascii=False) + "\n"
                             for (prompt, response) in zip(valid_prompts, valid_output)])
        # gathered_prompts_responses = [ppo_trainer.generate(
        #     [torch.LongTensor(self.tokenizer.encode(prompt)).to(self.current_device) for
        #      prompt in valid_prompts], return_prompt=False, generate_ref_response=False, **generation_kwargs)]
        # if self.accelerator.is_main_process:
        #     state_dict = self.accelerator.get_state_dict(self.model)
        #     self.model.save_pretrained(
        #         output_dir, state_dict=state_dict, safe_serialization=True
        #     )
        #     self.tokenizer.save_pretrained(output_dir)

    def log_stats(
            self,
            stats: dict,
            batch: dict,
            rewards: List[torch.FloatTensor],
            columns_to_log: typing.Iterable[str] = ("query", "response"),
    ):
        """
        A function that logs all the training stats. Call it at the end of each epoch.

        Args:
            stats (dict[str, Any]):
                A dictionary of training stats.
            batch (dict[str, Any]):
                A dictionary of batch data, this contains the queries and responses.
            rewards (`List[torch.FloatTensor]`):
                A tensor of rewards.
        """

        # all gather stats
        if not isinstance(rewards, torch.Tensor):
            rewards = torch.tensor(rewards).to(self.current_device)
        rewards = self.accelerator.gather(rewards).flatten()

        if self.config.log_with == "wandb":
            import wandb

            if any(column_to_log not in batch.keys() for column_to_log in columns_to_log):
                raise ValueError(f"Columns to log {columns_to_log} are not present in the batch {batch.keys()}.")

            batch_list = [batch[column_to_log] for column_to_log in columns_to_log]
            if self.is_distributed:
                gathered_batch_list = []
                for b in batch_list:
                    flattened = gather_object(b)
                    gathered_batch_list.append(flattened)
                batch_list = gathered_batch_list

        # Log only if we are in the main process
        if self.accelerator.is_main_process:
            logs = {}

            # Log stats
            if "query" not in batch.keys() and "response" not in batch.keys():
                # warn the user that the game logs will not be logged
                warnings.warn(
                    "The game logs will not be logged because the batch does not contain the keys 'query' and "
                    "'response'. "
                )
            elif self.config.log_with == "wandb":
                table_rows = [list(r) for r in zip(*batch_list, rewards.cpu().tolist())]
                logs.update({"game_log": wandb.Table(columns=[*columns_to_log, "reward"], rows=table_rows)})

            logs.update(stats)

            # manually cast in fp32 for bf16 torch tensors
            for k, v in logs.items():
                if isinstance(v, torch.Tensor) and v.dtype == torch.bfloat16:
                    logs[k] = v.float()

            logs["env/reward_mean"] = torch.mean(rewards).cpu().numpy().item()
            logs["env/reward_std"] = torch.std(rewards).cpu().numpy().item()
            logs["env/reward_dist"] = rewards.cpu().numpy()

            if self.config.log_with == "tensorboard":
                # 记录统计数据到 TensorBoard
                for key, value in logs.items():
                    if isinstance(value, (float, int)):  # 确保是数值类型
                        self.writer.add_scalar(key, value, self.current_step)

                # 更新步数
                self.current_step += 1

            self.accelerator.log(
                logs,
                step=self.current_step if self.config.log_with == "tensorboard" else None,
            )


def is_share_gpt_format(inputs: List):
    """
    [
        [{"role": "user", "content": "xx"}, {"role": "assistant", "content: "xxxx"}]
        [{"role": "user", "content": "yy"}, {"role": "assistant", "content: "yyyy"}]
        [{"role": "user", "content": "zz"}, {"role": "assistant", "content: "zzzz"}]
        ......
    ]
    """
    return isinstance(inputs[0], List) and \
        isinstance(inputs[0][0], dict) and \
        ("role" in inputs[0][0].keys() or "content" in inputs[0][0].keys())


def load_input(data_path: str, prompt_key: str, tokenizer, dataset_split=None):
    if data_path.endswith("json"):
        dataset = json.load(open(data_path, "r"))
        dataset = [row[prompt_key] for row in dataset] if prompt_key is not None else dataset
    elif data_path.endswith("jsonl"):
        dataset = [json.loads(row)[prompt_key] for row in open(data_path, "r").readlines()] if prompt_key is not None \
            else [json.loads(row) for row in open(data_path, "r").readlines()]
    else:
        # huggingface dataset or dataset dict object
        try:
            dataset = load_dataset(data_path)
        except ValueError:
            dataset = load_from_disk(data_path)
        if dataset_split is not None:
            dataset = dataset[dataset_split]
        dataset = dataset[prompt_key] if prompt_key is not None else prompt_key
    if not is_share_gpt_format(dataset):
        dataset = [[{"role": "user", "content": prompt}] for prompt in dataset]
    dataset = [prompt[:-1] if prompt[-1]['role'] == "assistant" else prompt for prompt in dataset]
    prompts_with_format = list(map(
        partial(tokenizer.apply_chat_template, tokenize=False, add_generation_prompt=True), dataset))
    return prompts_with_format


if __name__ == '__main__':
    parser = HfArgumentParser((ScriptArguments, PPOConfig))
    args, ppo_config = parser.parse_args_into_dataclasses()
    if not os.path.exists(args.output_dir):
        os.mkdir(args.output_dir)
    ppo_config.project_kwargs = {"logging_dir": args.output_dir}

    # We then define the arguments to pass to the sentiment analysis pipeline.
    # We set `return_all_scores` to True to get the sentiment score for each token.
    sent_kwargs = {"return_all_scores": True, "function_to_apply": "none", "batch_size": 16}

    trl_model_class = AutoModelForCausalLMWithValueHead if not args.use_seq2seq else AutoModelForSeq2SeqLMWithValueHead

    tokenizer = AutoTokenizer.from_pretrained(ppo_config.model_name)
    tokenizer.pad_token = tokenizer.eos_token

    # We retrieve the dataloader by calling the `build_dataset` function.
    # Compute that only on the main process for faster data processing.
    # see: https://github.com/huggingface/trl/pull/1255
    with PartialState().local_main_process_first():
        dataset = build_dataset(ppo_config.query_dataset, 4)  # ppo_config.dataset_num_proc)

    # set seed before initializing value head for deterministic eval
    set_seed(ppo_config.seed)

    # Now let's build the model, the reference model, and the tokenizer.
    if not args.use_peft:
        ref_model = trl_model_class.from_pretrained(
            ppo_config.model_name,
            trust_remote_code=args.trust_remote_code,
            torch_dtype=torch.bfloat16
        )
        device_map = None
        peft_config = None
    else:
        peft_config = LoraConfig(
            r=args.lora_r,
            lora_alpha=args.lora_alpha,
            bias="none",
            task_type="CAUSAL_LM",
        )
        ref_model = None
        # Copy the model to each device
        device_map = {"": Accelerator().local_process_index}

    model = trl_model_class.from_pretrained(
        ppo_config.model_name,
        trust_remote_code=args.trust_remote_code,
        device_map=device_map,
        peft_config=peft_config,
        torch_dtype=torch.bfloat16,
    )

    tokenizer = AutoTokenizer.from_pretrained(ppo_config.model_name)

    # Some tokenizers like GPT-2's don't have a padding token by default, so we set one here.
    tokenizer.pad_token_id = tokenizer.eos_token_id

    # We then build the PPOTrainer, passing the model, the reference model, the tokenizer
    ppo_trainer = CustomPPOTrainer(
        ppo_config,
        model,
        ref_model,
        tokenizer,
        query_dataset=dataset,
        sft_dataset=dataset,
        data_collator=collator,

    )

    # We then build the sentiment analysis pipeline, passing the model name and the
    # sentiment analysis pipeline arguments. Let's also make sure to set the device
    # to the same device as the PPOTrainer.
    device = ppo_trainer.accelerator.device
    if ppo_trainer.accelerator.num_processes == 1:
        if is_xpu_available():
            device = "xpu:0"
        elif is_npu_available():
            device = "npu:0"
        else:
            device = 0 if torch.cuda.is_available() else "cpu"  # to avoid a `pipeline` bug
    ds_plugin = ppo_trainer.accelerator.state.deepspeed_plugin
    if ds_plugin is not None and ds_plugin.is_zero3_init_enabled():
        with ds_plugin.zero3_init_context_manager(enable=False):
            teacher_dpo = AutoModelForCausalLM.from_pretrained(args.dpo_teacher, torch_dtype=torch.bfloat16).to(device)
            teacher_ref = AutoModelForCausalLM.from_pretrained(args.ref_teacher, torch_dtype=torch.bfloat16).to(device)
    else:
        teacher_dpo = AutoModelForCausalLM.from_pretrained(args.dpo_teacher, torch_dtype=torch.bfloat16).to(device)
        teacher_ref = AutoModelForCausalLM.from_pretrained(args.ref_teacher, torch_dtype=torch.bfloat16).to(device)
    # Some tokenizers like GPT-2's don't have a padding token by default, so we set one here.
    # if teacher_dpo.tokenizer.pad_token_id is None:
    #     sentiment_pipe.tokenizer.pad_token_id = tokenizer.pad_token_id
    #
    # if sentiment_pipe.model.config.pad_token_id is None:
    #     sentiment_pipe.model.config.pad_token_id = tokenizer.pad_token_id

    # We then define the arguments to pass to the `generate` function. These arguments
    # are passed to the `generate` function of the PPOTrainer, which is a wrapper around
    # the `generate` function of the trained model.
    generation_kwargs = {
        "min_length": -1,
        "top_k": 0.0,
        "top_p": 1.0,
        "do_sample": True,
        "pad_token_id": tokenizer.eos_token_id,
        "max_new_tokens": args.max_new_tokens,
    }
    valid_prompt = load_input(data_path=args.valid_dataset, tokenizer=tokenizer, prompt_key='chosen',
                              dataset_split="test") + \
                   load_input(data_path=args.test_dataset, tokenizer=tokenizer, prompt_key='instruction')
    already_steps = 0
    for epoch in range(args.num_train_epochs):
        for (batch, sft_batch) in tqdm(zip(ppo_trainer.dataloader, ppo_trainer.sft_dataloader)):
            query_tensors = batch["input_ids"]

            # Get response from gpt2
            response_tensors = ppo_trainer.generate(
                query_tensors, return_prompt=False, generate_ref_response=False, **generation_kwargs
            )
            batch["response"] = tokenizer.batch_decode(
                response_tensors)  # [(n, torch.any(torch.isnan(p))) for (n, p) in llm.named_parameters()]

            # prepare for the input
            query_tensor_list = [q.cpu().numpy().tolist() for q in query_tensors]
            response_tensor_list = [r.cpu().numpy().tolist() for r in response_tensors]
            input_ids = [q + r for (q, r) in zip(query_tensor_list, response_tensor_list)]
            max_length = max([len(x) for x in input_ids])
            input_ids = torch.LongTensor(
                [input_id + [tokenizer.pad_token_id] * (max_length - len(input_id)) for input_id in input_ids])
            labels = torch.LongTensor(
                [len(q) * [- 100] + r + [tokenizer.pad_token_id] * (max_length - len(q) - len(r)) for (q, r) in
                 zip(query_tensor_list, response_tensor_list)])
            # get reward
            dpo_logps_labels_all = []
            for j in range(ppo_config.batch_size // ppo_config.mini_batch_size):
                dpo_logps_labels_all.extend(get_logps_for_query_response(
                    teacher_dpo,
                    input_ids[j * ppo_config.mini_batch_size: (j + 1) * ppo_config.mini_batch_size],
                    labels[j * ppo_config.mini_batch_size: (j + 1) * ppo_config.mini_batch_size]))

            ref_logps_labels_all = []
            for j in range(ppo_config.batch_size // ppo_config.mini_batch_size):
                ref_logps_labels_all.extend(get_logps_for_query_response(
                    teacher_ref,
                    input_ids[j * ppo_config.mini_batch_size: (j + 1) * ppo_config.mini_batch_size],
                    labels[j * ppo_config.mini_batch_size: (j + 1) * ppo_config.mini_batch_size]))

            if args.reward_level == "token":
                rewards = [(dpoll.to(ppo_trainer.current_device) - refll.to(ppo_trainer.current_device))
                           for (dpoll, refll) in zip(dpo_logps_labels_all, ref_logps_labels_all)]
                rewards = [r.unsqueeze(0) if r.dim() == 0 else r for r in rewards]
            else:
                rewards = [(dpoll.to(ppo_trainer.current_device).sum() - refll.to(ppo_trainer.current_device).sum())
                           for (dpoll, refll) in zip(dpo_logps_labels_all, ref_logps_labels_all)]

            # Run PPO step
            stats = ppo_trainer.step(
                query_tensors, response_tensors, rewards,
                [{"input_ids": row[0], "labels": row[1]} for row in zip(sft_batch['input_ids'], sft_batch['labels'])])
            ppo_trainer.log_stats(stats, batch,
                                  torch.tensor([torch.sum(s) for s in rewards]).to(ppo_trainer.current_device),
                                  columns_to_log=["query", "response", "ref_response", "ref_rewards"])

            # step and save
            already_steps += 1
            if already_steps % args.save_steps == 0:
                ppo_trainer.save_model(os.path.join(args.output_dir, f"steps-{already_steps}"),
                                       valid_prompts=valid_prompt)

    ppo_trainer.save_model(os.path.join(args.output_dir, f"steps-end"), valid_prompts=valid_prompt)
    ppo_trainer.writer.close()

