import sys

sys.path.append("~/handbook-ADPA")
import datasets
from torch import nn, Tensor
from torch.utils.data import DataLoader
from transformers import is_datasets_available
from trl.trainer.utils import pad_to_length
from dataclasses import field, dataclass
from typing import Union, Dict, List, Literal, Tuple, Optional
import torch
import torch.nn.functional as F
from compress_logits import load_input_and_target_probs_fast
from scripts.run_simpo_ref import LPOTrainer
from scripts.run_precompute_logits import run_custom_dpo_main, CustomDPOConfig


@dataclass
class DistillDPOConfig(CustomDPOConfig):
    """
    the following parameters copied from alignment handbook
    """
    fkl_weight: Optional[float] = field(
        default=0,
        metadata={"help": "the loss weight of the fkl loss between teacher and student on chosen sequence"}
    )

    fkl_temperature: Optional[float] = field(
        default=1,
        metadata={"help": "the temperature parameter used to calculate the fkl loss "
                          "between teacher and student on chosen sequence"}
    )

    rkl_weight: Optional[float] = field(
        default=0,
        metadata={"help": "the loss weight of the rkl loss between teacher and student on rejected sequence"}
    )

    rkl_temperature: Optional[float] = field(
        default=1,
        metadata={"help": "the temperature parameter used to calculate the fkl loss "
                          "between teacher and student on chosen sequence"}
    )

    dpo_weight: Optional[float] = field(
        default=0,
        metadata={"help": "dpo weight loss"}
    )

    dpo_length_norm: Optional[bool] = field(
        default=True,
        metadata={"help": "whether normalize the length during dpo training"}
    )
    dpo_reference: Optional[str] = field(
        default="reference",
        metadata={"help": "use `initial` or `reference`"}
    )

    distillation_type: Optional[str] = field(
        default=None  # logsigmoid_beta, fkl_rkl, naive_chosen_fkl
    )

    distillation_from: Optional[str] = field(
        default="reference"
    )
    distillation_modify_wl: Optional[bool] = field(
        default=False  # logsigmoid_beta, fkl_rkl, naive_chosen_fkl
    )

    distillation_modify_wl_type: Optional[str] = field(
        default=None  # logsigmoid_beta, fkl_rkl, naive_chosen_fkl, random_rejected
    )
    rejected_distil_weight: Optional[float] = field(
        default=1
    )
    chosen_distil_weight: Optional[float] = field(
        default=1
    )

    distillation_weight: Optional[float] = field(
        default=0,
    )

    model_dpo_kl_alpha: Optional[float] = field(
        default=0,
    )

    use_margin_log_p: Optional[bool] = field(
        default=False,
        metadata={"help": "whether do pretraining on pretrained dataset"}
    )
    sft_on_pt: Optional[bool] = field(
        default=False,
        metadata={"help": "whether do pretraining on pretrained dataset"}
    )
    prob_stabilize: Optional[bool] = field(
        default=True,
        metadata={"help": "scale the probability into [+1e-8, -1e-8 + 1]"}
    )
    method: Optional[str] = field(
        default=None
    )
    teacher_prob_norm: Optional[bool] = field(
        default=None,
        metadata={"help": "Do extra operation on the teacher probability before distillation"}
    )


class DDPOTrainer(LPOTrainer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # Pass all other arguments using **kwargs
        self.gamma_increment = 0
        self.is_vision_model = False
        self.aux_loss_enabled = False
        # self.loss_type = "ipo"  # use `length` to normalize

    def get_train_dataloader(self) -> DataLoader:
        """
        Returns the training [`~torch.utils.data.DataLoader`].

        Will use no sampler if `train_dataset` does not implement `__len__`, a random sampler (adapted to distributed
        training if necessary) otherwise.

        Subclass and override this method if you want to inject some custom behavior.
        """
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")

        train_dataset = self.train_dataset
        data_collator = self.data_collator
        if is_datasets_available() and isinstance(train_dataset, datasets.Dataset):
            train_dataset = self._remove_unused_columns(train_dataset, description="training")
        else:
            data_collator = self._get_collator_with_removed_columns(data_collator, description="training")

        dataloader_params = {
            "batch_size": self._train_batch_size,
            "collate_fn": data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory}
        return self.accelerator.prepare(DataLoader(train_dataset, **dataloader_params))

    def gamma_update(self):
        self.gamma += self.gamma_increment

    def get_simpo_ref_loss(
            self,
            policy_chosen_logps: torch.FloatTensor,
            policy_rejected_logps: torch.FloatTensor,
            reference_chosen_logps: torch.FloatTensor,
            reference_rejected_logps: torch.FloatTensor,
            policy_chosen_length: torch.IntTensor,
            policy_rejected_length: torch.IntTensor,
            reference_chosen_length: torch.IntTensor,
            reference_rejected_length: torch.IntTensor,
    ) -> Tuple[
        torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:

        # step1: normalized by length
        if self.args.dpo_length_norm:
            policy_chosen_length = policy_chosen_length.to(policy_chosen_logps.device)
            policy_chosen_logps = policy_chosen_logps / policy_chosen_length
            policy_rejected_length = policy_rejected_length.to(policy_rejected_logps.device)
            policy_rejected_logps = policy_rejected_logps / policy_rejected_length
            reference_chosen_length = reference_chosen_length.to(reference_chosen_logps.device)
            reference_chosen_logps = reference_chosen_logps / reference_chosen_length
            reference_rejected_length = reference_rejected_length.to(reference_rejected_logps.device)
            reference_rejected_logps = reference_rejected_logps / reference_rejected_length

        # step2: call dpo loss
        pi_logratios = policy_chosen_logps - policy_rejected_logps
        if self.reference_free:
            ref_logratios = torch.tensor([0], dtype=pi_logratios.dtype, device=pi_logratios.device)
        else:
            ref_logratios = reference_chosen_logps - reference_rejected_logps

        pi_logratios = pi_logratios.to(self.accelerator.device)
        gamma_logratios = self.gamma / self.beta
        ref_logratios = ref_logratios.to(self.accelerator.device)
        logits = pi_logratios - ref_logratios - gamma_logratios

        # The beta is a temperature parameter for the DPO loss, typically something in the range of 0.1 to 0.5.
        # We ignore the reference model as beta -> 0. The label_smoothing parameter encodes our uncertainty about the labels and
        # calculates a conservative DPO loss.
        if self.loss_type == "sigmoid":
            losses = (
                    -F.logsigmoid(self.beta * logits) * (1 - self.label_smoothing)
                    - F.logsigmoid(-self.beta * logits) * self.label_smoothing
            )
        elif self.loss_type == "hinge":
            losses = torch.relu(1 - self.beta * logits)
        elif self.loss_type == "ipo":
            # eqn (17) of the paper where beta is the regularization parameter for the IPO loss, denoted by tau in the paper.
            losses = (logits - 1 / (2 * self.beta)) ** 2
        elif self.loss_type == "kto_pair":
            # eqn (7) of the HALOs paper
            chosen_KL = (policy_chosen_logps - reference_chosen_logps).mean().clamp(min=0)
            rejected_KL = (policy_rejected_logps - reference_rejected_logps).mean().clamp(min=0)

            chosen_logratios = policy_chosen_logps - reference_chosen_logps
            rejected_logratios = policy_rejected_logps - reference_rejected_logps
            # As described in the KTO report, the KL term for chosen (rejected) is estimated using the rejected (chosen) half.
            losses = torch.cat(
                (
                    1 - F.sigmoid(self.beta * (chosen_logratios - rejected_KL)),
                    1 - F.sigmoid(self.beta * (chosen_KL - rejected_logratios)),
                ),
                0,
            )
        else:
            raise ValueError(
                f"Unknown loss type: {self.loss_type}. Should be one of ['sigmoid', 'hinge', 'ipo', 'kto_pair']"
            )

        chosen_rewards = (
                self.beta
                * (
                        policy_chosen_logps.to(self.accelerator.device) - reference_chosen_logps.to(
                    self.accelerator.device)
                ).detach()
        )
        rejected_rewards = (
                self.beta
                * (
                        policy_rejected_logps.to(self.accelerator.device)
                        - reference_rejected_logps.to(self.accelerator.device)
                ).detach()
        )

        return losses, chosen_rewards, rejected_rewards, policy_chosen_logps, policy_rejected_logps, reference_chosen_logps, reference_rejected_logps

    def distillation_fkl_rkl(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            ref_chosen_logits_or_probs: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            ref_rejected_logits_or_probs: torch.Tensor,
            rejected_labels
    ):

        losses, fkl_loss, rkl_loss = 0.0, 0.0, 0.0
        if self.args.fkl_weight < 1e-6 and self.args.rkl_weight < 1e-6:
            return losses, fkl_loss, rkl_loss

        if self.args.fkl_weight > 1e-5:
            fkl_loss = self.get_kl_loss(
                target_logits_or_probs=ref_chosen_logits_or_probs.to(policy_chosen_logits_or_probs.device),
                input_logits_or_probs=policy_chosen_logits_or_probs,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            losses = losses + fkl_loss * self.args.fkl_weight

        if self.args.rkl_weight > 1e-5:
            rkl_loss = self.get_kl_loss(
                target_logits_or_probs=policy_rejected_logits_or_probs,
                input_logits_or_probs=ref_rejected_logits_or_probs.to(policy_chosen_logits_or_probs.device),
                labels=rejected_labels,
                distillation_temperature=self.args.rkl_temperature)
            losses = losses + rkl_loss * self.args.rkl_weight

        return losses, fkl_loss, rkl_loss

    @staticmethod
    def load_per_token_logps(
            logits: torch.FloatTensor,
            labels: torch.LongTensor,
            reference_per_token_logps,
            initial_per_token_logps=None,
            label_pad_token_id: int = -100,
    ):
        if logits.shape[:-1] != labels.shape:
            raise ValueError("Logits (batch and sequence length dim) and labels must have the same shape.")

        loss_mask = labels != label_pad_token_id

        # 使用非就地操作修改 labels
        labels = torch.where(labels == label_pad_token_id, torch.tensor(0).to(labels.device), labels)
        # torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)
        per_token_logps = torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)
        policy_per_token_logps = []
        initial_per_token_logps_ = []
        batch_size = logits.shape[0]

        for i in range(batch_size):
            policy_per_token_logps.append(per_token_logps[i][loss_mask[i]][: len(reference_per_token_logps[i])])
            if initial_per_token_logps is not None:
                initial_per_token_logps_.append(initial_per_token_logps[i][: len(reference_per_token_logps[i])])

        policy_per_token_logps = torch.cat(policy_per_token_logps)
        reference_per_token_logps = torch.cat([torch.tensor(j) for j in reference_per_token_logps])

        if initial_per_token_logps is not None:
            initial_per_token_logps = torch.cat([torch.tensor(j) for j in initial_per_token_logps_])

        return \
            policy_per_token_logps, \
                reference_per_token_logps.to(policy_per_token_logps.device), \
                initial_per_token_logps.to(
                    policy_per_token_logps.device) if initial_per_token_logps is not None else None

    def naive_chosen_rejected_fkl_distillation(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            ref_chosen_logits_or_probs: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            ref_rejected_logits_or_probs: torch.Tensor,
            rejected_labels
    ):
        distill_loss = 0
        if self.args.chosen_distil_weight > 1e-5:
            chosen_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_chosen_logits_or_probs,
                target_logits_or_probs=ref_chosen_logits_or_probs,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            distill_loss += chosen_fkl * self.args.chosen_distil_weight

        if self.args.rejected_distil_weight > 1e-5:
            rejected_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_rejected_logits_or_probs,
                target_logits_or_probs=ref_rejected_logits_or_probs,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            distill_loss += rejected_fkl * self.args.chosen_distil_weight

        return distill_loss

    def naive_chosen_rejected_rkl_distillation(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            ref_chosen_logits_or_probs: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            ref_rejected_logits_or_probs: torch.Tensor,
            rejected_labels
    ):
        distill_loss = 0
        if self.args.chosen_distil_weight > 1e-5:
            chosen_fkl = self.get_kl_loss(
                input_logits_or_probs=ref_chosen_logits_or_probs,
                target_logits_or_probs=policy_chosen_logits_or_probs,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            distill_loss += chosen_fkl * self.args.chosen_distil_weight

        if self.args.rejected_distil_weight > 1e-5:
            rejected_fkl = self.get_kl_loss(
                input_logits_or_probs=ref_rejected_logits_or_probs,
                target_logits_or_probs=policy_rejected_logits_or_probs,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            distill_loss += rejected_fkl * self.args.chosen_distil_weight

        return distill_loss

    def naive_chosen_rejected_ce_distillation(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            ref_chosen_logits_or_probs: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            ref_rejected_logits_or_probs: torch.Tensor,
            rejected_labels
    ):
        distill_loss = 0
        if self.args.chosen_distil_weight > 1e-5:
            chosen_ce = - self.get_negative_ce_loss(
                input_logits_or_probs=policy_chosen_logits_or_probs,
                target_logits_or_probs=ref_chosen_logits_or_probs,
                labels=chosen_labels
            )
            distill_loss += chosen_ce * self.args.chosen_distil_weight

        if self.args.rejected_distil_weight > 1e-5:
            rejected_ce = - self.get_negative_ce_loss(
                input_logits_or_probs=policy_rejected_logits_or_probs,
                target_logits_or_probs=ref_rejected_logits_or_probs,
                labels=rejected_labels
            )
            distill_loss += rejected_ce * self.args.rejected_distil_weight

        return distill_loss

    def value_reward_p_distillation(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            target_score_chosen: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            target_score_rejected: torch.Tensor,
            rejected_labels
    ):
        distillation_loss = 0
        if self.args.chosen_distil_weight > 0:
            chosen_negative_reward = self.get_value_reward_p(
                input_logits_or_probs=policy_chosen_logits_or_probs,
                target_score=target_score_chosen,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            distillation_loss += chosen_negative_reward * self.args.chosen_distil_weight

        if self.args.rejected_distil_weight > 0:
            rejected_negative_reward = self.get_value_reward_p(
                input_logits_or_probs=policy_rejected_logits_or_probs,
                target_score=target_score_rejected,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            distillation_loss += rejected_negative_reward * self.args.rejected_distil_weight

        return distillation_loss

    def value_reward_p_distillation_max_ent(
            self,
            policy_chosen_logits_or_probs: torch.Tensor,
            target_score_chosen: torch.Tensor,
            chosen_labels,
            policy_rejected_logits_or_probs: torch.Tensor,
            target_score_rejected: torch.Tensor,
            rejected_labels
    ):
        distillation_loss = 0
        if self.args.chosen_distil_weight > 0:
            chosen_negative_reward = self.get_value_reward_p_max_ent(
                input_logits_or_probs=policy_chosen_logits_or_probs,
                target_score=target_score_chosen,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            distillation_loss += chosen_negative_reward * self.args.chosen_distil_weight

        if self.args.rejected_distil_weight > 0:
            rejected_negative_reward = self.get_value_reward_p_max_ent(
                input_logits_or_probs=policy_rejected_logits_or_probs,
                target_score=target_score_rejected,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            distillation_loss += rejected_negative_reward * self.args.rejected_distil_weight

        return distillation_loss

    def get_kl_loss(self, input_logits_or_probs, target_logits_or_probs, labels, distillation_temperature=1):
        mask = (labels != -100).to(input_logits_or_probs.device)

        input_logits_or_probs[input_logits_or_probs < 1e-7] = 1e-7
        fkl = F.kl_div(
            input=input_logits_or_probs[mask].log(),
            target=target_logits_or_probs[mask],
            reduction='none'
        ).sum(-1).mean()
        return fkl
        # used to debug: (sum([i ==j for (i, j) in zip(input_logits_or_probs[0].argmax(-1), target_logits_or_probs[0].argmax(-1))]).item() / mask[0].sum()).item()

    def get_value_reward_p(self, input_logits_or_probs, target_score, labels, distillation_temperature=1):
        # 计算掩码，忽略标签为-100的token (sum([i == j for (i, j) in zip(input_logits[0].argmax(-1), target_logits[0].argmax(-1))]).item() / mask[0].sum()).item()
        mask = (labels != -100).to(input_logits_or_probs.device)
        input_logits_or_probs = input_logits_or_probs[mask]
        target_score = target_score[mask]
        value_reward_p = (input_logits_or_probs * target_score).sum(-1).mean()
        return - value_reward_p

    def get_value_reward_p_max_ent(self, input_logits_or_probs, target_score, labels, distillation_temperature=1):
        mask = (labels != -100).to(input_logits_or_probs.device)
        input_logits_or_probs = input_logits_or_probs[mask]  # batch_size * seq_len, vocab_size
        target_score = target_score[mask]
        value_reward_p = self.beta * torch.log((input_logits_or_probs * torch.exp(target_score / self.beta)).sum(-1)).mean()
        return - value_reward_p

    def get_negative_ce_loss(self, input_logits_or_probs, target_logits_or_probs, labels, distillation_temperature=1):
        # 计算掩码，忽略标签为-100的token (sum([i == j for (i, j) in zip(input_logits[0].argmax(-1), target_logits[0].argmax(-1))]).item() / mask[0].sum()).item()
        mask = (labels != -100).to(input_logits_or_probs.device)
        input_logits_or_probs = input_logits_or_probs[mask]
        input_logits_or_probs[input_logits_or_probs < 1e-7] = 1e-7
        target_logits_or_probs = target_logits_or_probs[mask]
        negative_ce = (target_logits_or_probs * torch.log(input_logits_or_probs)).sum(-1).mean()
        return negative_ce

    @staticmethod
    def weighted_reward(input_per_token_logps, target_per_token_logps, weight):
        """
        input_per_token_logps: [length]
        target_per_token_logps: [length]
        weight: [length]
        """
        return ((input_per_token_logps - target_per_token_logps) * weight).sum() / weight.sum()

    def margin_softmax_reward(self, input_per_token_logps, reference_per_token_logps, initial_per_token_logps):
        weight = F.sigmoid(reference_per_token_logps - initial_per_token_logps)
        return self.weighted_reward(input_per_token_logps, reference_per_token_logps,
                                    weight=weight)

    def negative_margin_softmax_reward(self, input_per_token_logps, reference_per_token_logps, initial_per_token_logps):
        weight = F.sigmoid(- reference_per_token_logps + initial_per_token_logps)
        return self.weighted_reward(input_per_token_logps, reference_per_token_logps,
                                    weight=weight)

    def dpkd_reward(self, input_per_token_logps, target_per_token_logps):
        return self.weighted_reward(input_per_token_logps, target_per_token_logps,
                                    weight=torch.ones_like(input_per_token_logps))

    def dpo_reward(self, input_per_token_logps, target_per_token_logps):
        return self.dpkd_reward(input_per_token_logps, target_per_token_logps) * len(input_per_token_logps)

    def tpkd_reward(self, input_per_token_logps, target_per_token_logps):
        return self.weighted_reward(input_per_token_logps, target_per_token_logps,
                                    weight=torch.exp(target_per_token_logps))

    def negative_tpkd_reward(self, input_per_token_logps, target_per_token_logps):
        return self.weighted_reward(input_per_token_logps, target_per_token_logps,
                                    weight=1 - torch.exp(target_per_token_logps))

    def spkd_reward(self, input_per_token_logps, target_per_token_logps):
        return self.weighted_reward(input_per_token_logps, target_per_token_logps,
                                    weight=torch.exp(input_per_token_logps))

    def negative_spkd_reward(self, input_per_token_logps, target_per_token_logps):
        return self.weighted_reward(input_per_token_logps, target_per_token_logps,
                                    weight=1 - torch.exp(input_per_token_logps))

    @staticmethod
    def concatenated_inputs(
            batch: Dict[str, Union[List, torch.LongTensor]],
            is_encoder_decoder: bool = False,
            is_vision_model: bool = False,
            label_pad_token_id: int = -100,
            padding_value: int = 0,
            device: Optional[torch.device] = None,
    ) -> dict[str, Tensor]:
        """Concatenate the chosen and rejected inputs into a single tensor.

        Args:
            batch: A batch of data. Must contain the keys 'chosen_input_ids' and 'rejected_input_ids', which are tensors of shape (batch_size, sequence_length).
            is_encoder_decoder: Whether the model is an encoder-decoder model.
            label_pad_token_id: The label pad token id.
            padding_value: The padding value to use for the concatenated inputs_ids.
            device: The device for the concatenated inputs.

        Returns:
            A dictionary containing the concatenated inputs under the key 'concatenated_input_ids'.
        """
        concatenated_batch = {}

        if is_encoder_decoder:
            max_length = max(batch["chosen_labels"].shape[1], batch["rejected_labels"].shape[1])
        else:
            max_length = max(batch["chosen_input_ids"].shape[1], batch["rejected_input_ids"].shape[1])

        for k in batch:
            if k.startswith("chosen") and isinstance(batch[k], torch.Tensor):
                if "labels" in k or is_encoder_decoder:
                    pad_value = label_pad_token_id
                elif k.endswith("_input_ids"):
                    pad_value = padding_value
                elif k.endswith("_attention_mask"):
                    pad_value = 0
                concatenated_key = k.replace("chosen", "concatenated")
                concatenated_batch[concatenated_key] = pad_to_length(batch[k], max_length, pad_value=pad_value)
        for k in batch:
            if k.startswith("rejected") and isinstance(batch[k], torch.Tensor):
                if "labels" in k or is_encoder_decoder:
                    pad_value = label_pad_token_id
                elif k.endswith("_input_ids"):
                    pad_value = padding_value
                elif k.endswith("_attention_mask"):
                    pad_value = 0
                concatenated_key = k.replace("rejected", "concatenated")
                concatenated_batch[concatenated_key] = torch.cat(
                    (
                        concatenated_batch[concatenated_key],
                        pad_to_length(batch[k], max_length, pad_value=pad_value),
                    ),
                    dim=0,
                ).to(device=device)

        if is_encoder_decoder:
            concatenated_batch["concatenated_input_ids"] = batch["prompt_input_ids"].repeat(2, 1).to(device=device)
            concatenated_batch["concatenated_attention_mask"] = (
                batch["prompt_attention_mask"].repeat(2, 1).to(device=device)
            )

        if is_vision_model:
            concatenated_batch["pixel_values"] = batch["prompt_pixel_values"].repeat(2, 1, 1, 1, 1).to(device=device)
            concatenated_batch["pixel_attention_mask"] = (
                batch["prompt_pixel_attention_mask"].repeat(2, 1, 1, 1).to(device=device)
            )
        return concatenated_batch

    def concatenated_forward(
            self, model: nn.Module, batch: Dict[str, Union[List, torch.LongTensor]]
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:
        """Run the given model on the given batch of inputs, concatenating the chosen and rejected inputs together.

        We do this to avoid doing two forward passes, because it's faster for FSDP.
        """
        concatenated_batch = self.concatenated_inputs(
            batch,
            is_encoder_decoder=self.is_encoder_decoder,
            is_vision_model=self.is_vision_model,
            label_pad_token_id=self.label_pad_token_id,
            padding_value=self.padding_value,
            device=self.accelerator.device,
        )
        len_chosen = batch["chosen_labels"].shape[0]

        model_kwargs = {}

        if self.is_encoder_decoder:
            model_kwargs["labels"] = concatenated_batch["concatenated_labels"]
            model_kwargs["decoder_input_ids"] = concatenated_batch.pop("concatenated_decoder_input_ids", None)

        if self.is_vision_model:
            model_kwargs["pixel_values"] = concatenated_batch["pixel_values"]
            model_kwargs["pixel_attention_mask"] = concatenated_batch["pixel_attention_mask"]

        if self.aux_loss_enabled:
            model_kwargs["output_router_logits"] = True

        outputs = model(
            concatenated_batch["concatenated_input_ids"],
            attention_mask=concatenated_batch["concatenated_attention_mask"],
            use_cache=False,
            **model_kwargs,
        )
        all_logits = outputs.logits

        all_logps, size_completion = self.get_batch_logps(
            all_logits,
            concatenated_batch["concatenated_labels"],
            # average_log_prob=self.loss_type == "ipo",
            is_encoder_decoder=self.is_encoder_decoder,
            label_pad_token_id=self.label_pad_token_id,
        )

        def cross_entropy_loss(logits, labels):
            if not self.is_encoder_decoder:
                # Shift so that tokens < n predict n
                logits = logits[..., :-1, :].contiguous()
                labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = nn.CrossEntropyLoss()
            logits = logits.view(-1, logits.shape[-1])
            labels = labels.view(-1)
            # Enable model parallelism
            labels = labels.to(logits.device)
            loss = loss_fct(logits, labels)
            return loss

        labels = concatenated_batch["concatenated_labels"].clone()
        nll_loss = cross_entropy_loss(all_logits[:len_chosen], labels[:len_chosen])
        if self.loss_type == "ipo":
            all_logps = all_logps / size_completion

        chosen_logps = all_logps[:len_chosen]
        rejected_logps = all_logps[len_chosen:]

        chosen_logits = all_logits[:len_chosen]
        rejected_logits = all_logits[len_chosen:]

        if self.aux_loss_enabled:
            return (chosen_logps, rejected_logps, chosen_logits, rejected_logits, nll_loss,
                    outputs.aux_loss), concatenated_batch

        return (chosen_logps, rejected_logps, chosen_logits, rejected_logits, nll_loss), concatenated_batch

    def load_margin_logp_for_labels(self, margin_logp_every, labels):
        rst = []
        for mle_batch, label_batch in zip(margin_logp_every, labels):
            left = torch.where(label_batch != -100)[0].min().item()
            for label, mle in zip(label_batch[left:], mle_batch):
                label = label.item()
                if label not in mle['indices']:
                    rst.append(0)
                else:
                    label_index = [i for (i, vocab_idx) in enumerate(mle['indices']) if vocab_idx == label][0]
                    rst.append(mle['values'][label_index])
        return torch.tensor(rst)

    def get_batch_loss_metrics(
            self,
            model,
            batch: Dict[str, Union[List, torch.LongTensor]],
            train_eval: Literal["train", "eval"] = "train",
    ):
        """Compute the DPO loss and other metrics for the given batch of inputs for train or test."""
        metrics = {}
        prefix = "eval_" if train_eval == "eval" else ""
        losses = 0

        # SFT LOSS
        forward_output, concatenated_batch = self.concatenated_forward(model, batch)
        (
            policy_chosen_logps,
            policy_rejected_logps,
            policy_chosen_logits,
            policy_rejected_logits,
            policy_nll_loss,
        ) = forward_output[:5]
        if self.aux_loss_enabled:
            aux_loss = forward_output[5]
        metrics[f"{prefix}loss/sft"] = policy_nll_loss.detach().mean().cpu().item()
        sft_loss = policy_nll_loss

        if self.args.sft_on_chosen > 1e-5:
            losses = losses + self.args.sft_on_chosen * sft_loss

        # shift the label and logits
        policy_chosen_logits = policy_chosen_logits[:, :-1,
                               :]  # policy_chosen_logits[0][:1255, :][batch['chosen_labels'][0] != -100].max(-1)
        policy_rejected_logits = policy_rejected_logits[:, :-1, :]
        batch_size, seq_len, vocab_size = policy_chosen_logits.shape
        chosen_labels = concatenated_batch['concatenated_labels'][:batch_size][:, 1:]
        rejected_labels = concatenated_batch['concatenated_labels'][batch_size:][:, 1:]
        # calculate the loss
        labels = concatenated_batch["concatenated_labels"][:, 1:].clone()
        length_all = (labels != self.label_pad_token_id).sum(-1)
        policy_chosen_length = length_all[:batch_size]
        policy_rejected_length = length_all[batch_size:]

        if self.args.sft_on_pt > 1e-5 and "pt_input_ids" in batch.keys():
            input_ids = batch["pt_input_ids"]
            pt_logits = model(input_ids=batch["pt_input_ids"]).logits
            # 我们只需要计算目标 token 的概率
            shifted_logits = pt_logits[:, :-1, :]  # 忽略最后一个 logit，因为没有后续 token 作为目标
            shifted_input_ids = input_ids[:, 1:]  # 忽略第一个输入 token，因为没有先前 token 作为上下文
            log_probs = F.log_softmax(shifted_logits, dim=-1)  # 对每个 logit 取 softmax 再取对数
            # 获取每个 target token 的 log probability
            target_log_probs = log_probs.gather(2, shifted_input_ids.unsqueeze(-1)).squeeze(-1)
            pt_loss = - target_log_probs.mean()
            losses += self.args.sft_on_pt * pt_loss
            metrics[f"{prefix}loss/pt-loss"] = pt_loss.detach().mean().cpu().item()

        metrics[f"{prefix}logits/rejected"] = policy_rejected_logits.detach().mean().cpu()
        metrics[f"{prefix}logits/chosen"] = policy_chosen_logits.detach().mean().cpu()

        if not (
                self.args.dpo_weight > 1e-5 or self.args.distillation_weight > 1e-5 or train_eval == "eval"):
            return losses.mean(), metrics

        # 获取 policy 和 reference 的 chosen 和 rejected 的 logps 和 logits / probs
        reference_chosen_logps, reference_rejected_logps = None, None
        policy_chosen_logits_probs, policy_rejected_logits_probs = None, None
        if f"{self.args.dpo_reference}_chosen_logps" in batch and f"{self.args.dpo_reference}_rejected_logps" in batch:
            reference_chosen_logps = batch[f"{self.args.dpo_reference}_chosen_logps"].squeeze(1)
            reference_rejected_logps = batch[f"{self.args.dpo_reference}_rejected_logps"].squeeze(1)

        # pLaD
        # L_{rank} ^ {cal} = max(0, \beta - logP_\theta(y_ + | x) + logP_\theta(y_ - | x))
        if self.args.method == "plad":
            relative_logp = policy_chosen_logps - policy_rejected_logps - self.beta
            losses = F.relu(- relative_logp)

        elif self.args.method == "ddpo":
            reference_chosen_logps = batch[f"initial_chosen_logps"].squeeze(1).to(policy_chosen_logps.device)
            reference_rejected_logps = batch[f"initial_rejected_logps"].squeeze(1).to(policy_chosen_logps.device)
            teacher_pol_chosen_logps = batch[f"reference_chosen_logps"].squeeze(1)
            teacher_pol_rejected_logps = batch[f"reference_rejected_logps"].squeeze(1)
            teacher_ref_chosen_logps = batch[f"teacher_ref_chosen_logps"].squeeze(1)
            teacher_ref_rejected_logps = batch[f"teacher_ref_rejected_logps"].squeeze(1)
            rxy1 = (teacher_pol_chosen_logps - teacher_ref_chosen_logps).to(policy_chosen_logps.device)
            rxy2 = (teacher_pol_rejected_logps - teacher_ref_rejected_logps).to(policy_chosen_logps.device)
            stu_margin = (policy_chosen_logps + reference_rejected_logps - policy_rejected_logps - reference_chosen_logps) * self.beta
            tea_margin = (rxy1 - rxy2) * self.beta
            metrics[f"{prefix}ddpo/stu_margin"] = stu_margin.detach().mean().cpu()
            metrics[f"{prefix}ddpo/tea_margin"] = tea_margin.detach().mean().cpu()
            losses = (tea_margin -stu_margin) ** 2  # MSE loss defined in DDPO

        elif self.args.method == "atkd":
            policy_chosen_logits_probs = F.softmax(policy_chosen_logits, dim=-1)
            compressed_probs: List[List[Dict]] = [batch[f"{self.args.distillation_from}_chosen_probs"][i][0] for i in range(batch_size)]
            input_logits_or_probs, target_logits_or_probs, is_easy_mask = load_input_and_target_probs_fast(
                compressed_probs=compressed_probs,
                input_probs=policy_chosen_logits_probs,
                labels=chosen_labels,
                prob_stabilize=True,
                atkd=True,
            )  # according to the ATKD paper, 50% tokens will be set as hard token, while others set as easy token.
            mask = (chosen_labels != -100).to(input_logits_or_probs.device)
            fkl = F.kl_div(
                input=input_logits_or_probs.log(),
                target=target_logits_or_probs,
                reduction='none'
            )  # batch_size, seq_len, mini_vocab
            fkl = self.beta * fkl * is_easy_mask + (1 - self.beta) * fkl * (~is_easy_mask) # beta set as 0.2 in the paper
            fkl = fkl[mask].sum(-1).mean()
            losses += self.args.distillation_weight * fkl

        if self.args.distillation_weight > 1e-5 and self.args.distillation_type is not None:
            reference_chosen_logits_probs = None
            reference_rejected_logits_probs = None

            # Chosen
            if self.args.chosen_distil_weight > 0:
                policy_chosen_logits_probs = F.softmax(policy_chosen_logits, dim=-1)
                if self.args.distillation_modify_wl and self.args.distillation_modify_wl_type in ["all", "chosen"]:
                    from_chosen_or_rejected = "chosen"
                else:
                    from_chosen_or_rejected = None
                if self.args.use_margin_log_p:
                    compressed_probs = batch["chosen_margin_logp_every"]
                else:
                    compressed_probs: List[List[Dict]] = [batch[f"{self.args.distillation_from}_chosen_probs"][i][0] for i in
                                                              range(batch_size)]
                policy_chosen_logits_probs, reference_chosen_logits_probs = load_input_and_target_probs_fast(
                    compressed_probs=compressed_probs,
                    input_probs=policy_chosen_logits_probs,
                    labels=chosen_labels,
                    from_chosen_or_rejected=from_chosen_or_rejected,
                    prob_stabilize=self.args.prob_stabilize,
                )
                if self.args.teacher_prob_norm == "softmax":
                    reference_chosen_logits_probs = F.softmax(reference_chosen_logits_probs, dim=-1)
                elif self.args.teacher_prob_norm == "argmax":
                    reference_chosen_logits_probs = torch.argmax(reference_chosen_logits_probs, dim=-1)
                    reference_chosen_logits_probs = torch.nn.functional.one_hot(
                        reference_chosen_logits_probs, num_classes=reference_chosen_logits_probs.shape[-1]).float()

            # Rejected
            if self.args.rejected_distil_weight > 0:
                policy_rejected_logits_probs = F.softmax(policy_rejected_logits, dim=-1)
                if self.args.distillation_modify_wl and self.args.distillation_modify_wl_type in ["all", "rejected"]:
                    from_chosen_or_rejected = "rejected"
                else:
                    from_chosen_or_rejected = None

                if self.args.use_margin_log_p:
                    compressed_probs = batch["rejected_margin_logp_every"]
                else:
                    compressed_probs = [batch[f"{self.args.distillation_from}_rejected_probs"][i][0] for i in range(batch_size)]

                policy_rejected_logits_probs, reference_rejected_logits_probs = load_input_and_target_probs_fast(
                    compressed_probs=compressed_probs,
                    input_probs=policy_rejected_logits_probs,
                    labels=rejected_labels,
                    from_chosen_or_rejected=from_chosen_or_rejected,
                    prob_stabilize=self.args.prob_stabilize,
                )

                if self.args.teacher_prob_norm == "softmax":
                    reference_rejected_logits_probs = F.softmax(reference_rejected_logits_probs, dim=-1)
                elif self.args.teacher_prob_norm == "argmax":
                    reference_rejected_logits_probs = torch.argmax(reference_rejected_logits_probs, dim=-1)
                    reference_rejected_logits_probs = torch.nn.functional.one_hot(
                        reference_rejected_logits_probs, num_classes=policy_rejected_logits_probs.shape[-1]).float()

            if self.args.distillation_type == "naive_chosen_fkl":
                distillation_loss = self.naive_chosen_fkl_distillation(
                    policy_chosen_logits_or_probs=policy_chosen_logits_probs,
                    ref_chosen_logits_or_probs=reference_chosen_logits_probs,
                    chosen_labels=chosen_labels,
                    policy_rejected_logits_or_probs=policy_rejected_logits_probs,
                    ref_rejected_logits_or_probs=reference_rejected_logits_probs,
                    rejected_labels=rejected_labels
                )

            elif self.args.distillation_type == "naive_chosen_rejected_fkl":
                distillation_loss = self.naive_chosen_rejected_fkl_distillation(
                    policy_chosen_logits_or_probs=policy_chosen_logits_probs,
                    ref_chosen_logits_or_probs=reference_chosen_logits_probs,
                    chosen_labels=chosen_labels,
                    policy_rejected_logits_or_probs=policy_rejected_logits_probs,
                    ref_rejected_logits_or_probs=reference_rejected_logits_probs,
                    rejected_labels=rejected_labels
                )

            elif self.args.distillation_type == "naive_chosen_rejected_ce":
                distillation_loss = self.naive_chosen_rejected_ce_distillation(
                    policy_chosen_logits_or_probs=policy_chosen_logits_probs,
                    ref_chosen_logits_or_probs=reference_chosen_logits_probs,
                    chosen_labels=chosen_labels,
                    policy_rejected_logits_or_probs=policy_rejected_logits_probs,
                    ref_rejected_logits_or_probs=reference_rejected_logits_probs,
                    rejected_labels=rejected_labels
                )

            elif self.args.distillation_type == "value_reward_p":
                distillation_loss = self.value_reward_p_distillation(
                    policy_chosen_logits_or_probs=policy_chosen_logits_probs,
                    target_score_chosen=reference_chosen_logits_probs,
                    chosen_labels=chosen_labels,
                    policy_rejected_logits_or_probs=policy_rejected_logits_probs,
                    target_score_rejected=reference_rejected_logits_probs,
                    rejected_labels=rejected_labels
                )
            elif self.args.distillation_type == "value_reward_p_max_ent":
                distillation_loss = self.value_reward_p_distillation_max_ent(
                    policy_chosen_logits_or_probs=policy_chosen_logits_probs,
                    target_score_chosen=reference_chosen_logits_probs,
                    chosen_labels=chosen_labels,
                    policy_rejected_logits_or_probs=policy_rejected_logits_probs,
                    target_score_rejected=reference_rejected_logits_probs,
                    rejected_labels=rejected_labels
                )
            if distillation_loss != 0 and not isinstance(distillation_loss, int):
                metrics[f"{prefix}loss/distillation_loss"] = distillation_loss.item()
                losses = losses + self.args.distillation_weight * distillation_loss.mean()

            if train_eval == "eval" and reference_chosen_logits_probs is not None and reference_rejected_logits_probs is not None:
                reference_chosen_logits_probs = reference_chosen_logits_probs.to(policy_chosen_logits_probs.device)
                reference_rejected_logits_probs = reference_rejected_logits_probs.to(
                    policy_rejected_logits_probs.device)
                metrics[f"{prefix}loss/fkl_chosen"] = self.get_kl_loss(
                    input_logits_or_probs=policy_chosen_logits_probs,
                    target_logits_or_probs=reference_chosen_logits_probs,
                    labels=chosen_labels)
                metrics[f"{prefix}loss/fkl_rejected"] = self.get_kl_loss(
                    input_logits_or_probs=policy_rejected_logits_probs,
                    target_logits_or_probs=reference_rejected_logits_probs,
                    labels=rejected_labels)

        if (self.args.dpo_weight > 1e-5 or train_eval == "eval") and reference_chosen_logps is not None:
            if f"{self.args.dpo_reference}_chosen_length" in batch:
                reference_chosen_length = torch.IntTensor(batch[f"{self.args.dpo_reference}_chosen_length"]).squeeze()
            else:
                reference_chosen_length = torch.IntTensor([len(batch["reference_chosen_probs"][i][0]) for i in range(batch_size)]).squeeze()
            if f"{self.args.dpo_reference}_rejected_length" in batch:
                reference_rejected_length = torch.IntTensor(batch[f"{self.args.dpo_reference}_rejected_length"]).squeeze()
            else:
                reference_rejected_length = torch.IntTensor([len(batch["reference_rejected_probs"][i][0]) for i in range(batch_size)]).squeeze()
            dpo_losses, chosen_rewards, rejected_rewards, _, _, _, _ = \
                self.get_simpo_ref_loss(
                    policy_chosen_logps=policy_chosen_logps,
                    policy_rejected_logps=policy_rejected_logps,
                    reference_chosen_logps=reference_chosen_logps,
                    reference_rejected_logps=reference_rejected_logps,
                    policy_chosen_length=policy_chosen_length,
                    policy_rejected_length=policy_rejected_length,
                    reference_chosen_length=reference_chosen_length,
                    reference_rejected_length=reference_rejected_length
                )
            if self.args.dpo_weight > 1e-5:
                losses = losses + self.args.dpo_weight * dpo_losses.mean()
            reward_accuracies = (chosen_rewards > rejected_rewards).float()
            metrics[f"{prefix}rewards/rejected"] = rejected_rewards.mean().cpu()
            metrics[f"{prefix}rewards/accuracies"] = reward_accuracies.mean().cpu()
            metrics[f"{prefix}rewards/margins"] = (chosen_rewards - rejected_rewards).mean().cpu()
            metrics[f"{prefix}rewards/margins_gamma"] = (chosen_rewards - rejected_rewards - self.gamma).mean().cpu()
            metrics[f"{prefix}rewards/margins_nobeta"] = ((chosen_rewards - rejected_rewards) / self.beta).mean().cpu()
            metrics[f"{prefix}length/chosen"] = policy_chosen_length.to(torch.float).mean().cpu()
            metrics[f"{prefix}length/rejected"] = policy_rejected_length.to(torch.float).mean().cpu()
        metrics[f"{prefix}logps_rejected"] = policy_rejected_logps.detach().mean().cpu()
        metrics[f"{prefix}logps_chosen"] = policy_chosen_logps.detach().mean().cpu()
        metrics[f"{prefix}logps_avg/rejected"] = (policy_rejected_logps / policy_rejected_length).detach().mean().cpu()
        metrics[f"{prefix}logps_avg/chosen"] = (policy_chosen_logps / policy_chosen_length).detach().mean().cpu()

        # teacher chosen fkl
        if train_eval == "eval" and 'reference_chosen_probs' in batch.keys():
            policy_chosen_probs = F.softmax(policy_chosen_logits, dim=-1)
            policy_chosen_probs, reference_chosen_probs = load_input_and_target_probs_fast(
                compressed_probs=[batch["reference_chosen_probs"][i][0] for i in range(batch_size)],
                input_probs=policy_chosen_probs,
                labels=chosen_labels)
            chosen_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_chosen_probs,
                target_logits_or_probs=reference_chosen_probs,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            metrics[f"{prefix}chosen_fkl_teacher"] = (chosen_fkl).detach().mean().cpu()

        # teacher rejected fkl
        if train_eval == "eval" and 'reference_rejected_probs' in batch.keys():
            policy_rejected_probs = F.softmax(policy_rejected_logits, dim=-1)
            policy_rejected_probs, reference_rejected_probs = load_input_and_target_probs_fast(
                compressed_probs=[batch["reference_rejected_probs"][i][0] for i in range(batch_size)],
                input_probs=policy_rejected_probs,
                labels=rejected_labels)
            rejected_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_rejected_probs,
                target_logits_or_probs=reference_rejected_probs,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            metrics[f"{prefix}rejected_fkl_teacher"] = (rejected_fkl).detach().mean().cpu()

        # sft teacher chosen fkl
        if train_eval == "eval" and 'teacher_ref_chosen_probs' in batch.keys():
            policy_chosen_probs = F.softmax(policy_chosen_logits, dim=-1)
            policy_chosen_probs, reference_chosen_probs = load_input_and_target_probs_fast(
                compressed_probs=[batch["teacher_ref_chosen_probs"][i][0] for i in range(batch_size)],
                input_probs=policy_chosen_probs,
                labels=chosen_labels)
            chosen_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_chosen_probs,
                target_logits_or_probs=reference_chosen_probs,
                labels=chosen_labels,
                distillation_temperature=self.args.fkl_temperature)
            metrics[f"{prefix}chosen_fkl_ref"] = (chosen_fkl).detach().mean().cpu()

        # sft teacher chosen fkl
        if train_eval == "eval" and 'teacher_ref_rejected_probs' in batch.keys():
            policy_rejected_probs = F.softmax(policy_rejected_logits, dim=-1)
            policy_rejected_probs, reference_rejected_probs = load_input_and_target_probs_fast(
                compressed_probs=[batch["teacher_ref_rejected_probs"][i][0] for i in range(batch_size)],
                input_probs=policy_rejected_probs,
                labels=rejected_labels)
            rejected_fkl = self.get_kl_loss(
                input_logits_or_probs=policy_rejected_probs,
                target_logits_or_probs=reference_rejected_probs,
                labels=rejected_labels,
                distillation_temperature=self.args.fkl_temperature)
            metrics[f"{prefix}rejected_fkl_ref"] = (rejected_fkl).detach().mean().cpu()

        return losses.mean(), metrics


def lpo_training_main():
    run_custom_dpo_main(DistillDPOConfig, DDPOTrainer)


if __name__ == '__main__':
    torch.autograd.set_detect_anomaly(True)
    lpo_training_main()
