import sys

from jinja2 import TemplateError

sys.path.append("/GLOBALFS/sysu_xjquan_2/spgao/alignment-handbook-cs")
import inspect
import logging
import random
import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Literal, Optional, Tuple, Union, Any

import datasets
import torch
import torch.nn as nn
import transformers
from accelerate import PartialState
from accelerate.utils import is_deepspeed_available
from datasets import Dataset
from datasets import DatasetDict
from peft import PeftConfig
from peft import PeftModel, prepare_model_for_kbit_training, get_peft_model
from torch import Tensor
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import (
    AutoModelForCausalLM,
    DataCollator,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    Trainer,
    TrainingArguments, is_wandb_available,
)
from transformers import set_seed
from transformers.trainer_callback import TrainerCallback
from transformers.trainer_utils import EvalLoopOutput
from trl import DPOConfig
from trl import DPOTrainer, is_peft_available
from trl.trainer import peft_module_casting_to_bf16
from trl.trainer.utils import DPODataCollatorWithPadding, disable_dropout_in_model

from alignment import (
    DataArguments,
    H4ArgumentParser,
    ModelArguments,
    apply_chat_template,
    get_checkpoint,
    get_datasets,
    get_kbit_device_map,
    get_peft_config,
    get_quantization_config,
    get_tokenizer,
    is_adapter_model,
)
from compress_logits import compress_logits, compress_probs

logger = logging.getLogger(__name__)
if is_deepspeed_available():
    pass


def replace_last_eos_with_pad(tensor, eos_token_id, pad_token_id):
    # 获取tensor的尺寸
    batch_size, seq_len = tensor.size()

    # 创建一个与tensor相同尺寸的mask，并初始化为False
    mask = torch.zeros_like(tensor, dtype=torch.bool)

    # 遍历每个batch
    for i in range(batch_size):
        # 找到当前batch中最后一个eos_token_id的位置
        last_eos_index = (tensor[i] == eos_token_id).nonzero(as_tuple=True)[0].max()
        # 将该位置的mask设置为True
        mask[i, last_eos_index] = True

    # 用pad_token_id替换mask为True的位置
    tensor[mask] = pad_token_id

    return tensor


def process_response(r: str, assistant_start: str = None, assistant_end: str = None):
    rst = r
    if assistant_start is not None and r.startswith(assistant_start):
        rst = rst[len(assistant_start):]
    if assistant_end is not None and r.endswith(assistant_end):
        rst = rst[: -len(assistant_end)]
    return rst


@dataclass
class CustomDPOConfig(DPOConfig):
    """
    the following parameters copied from alignment handbook
    """
    ref_model: Optional[str] = field(
        default=None,
    )
    beta: Optional[float] = field(
        default=0.1,
        metadata={"help": "The beta factor in DPO loss. Higher beta means less divergence from the initial policy."},
    )
    hub_model_revision: Optional[str] = field(
        default="main",
        metadata={"help": "The Hub model branch to push the model to."},
    )
    logging_first_step: bool = field(
        default=True,
        metadata={"help": "Whether to log and evaluate the first global_step or not."},
    )
    max_prompt_length: Optional[int] = field(
        default=None,
        metadata={"help": "For DPO, the maximum length of the prompt to use for conditioning the model."},
    )
    max_length: Optional[int] = field(
        default=None,
        metadata={"help": "Used by TRL for reward model training, which tries to read this parameter in init."},
    )
    optim: Optional[str] = field(default="rmsprop")
    remove_unused_columns: bool = field(default=False)
    loss_type: Optional[str] = field(default="sigmoid", metadata={"help": "The loss type for DPO."})

    """
    接下来的参数控制是在超算上跑，还是在实验室的服务器上跑；
    """
    compute_on_8a100: Optional[bool] = field(
        default=True,
        metadata={"help": "在超算上跑的时候，设置数据并行，并且不需要提前计算 reference log p"}
    )
    pre_compute_reference_dataset: Optional[str] = field(
        default=None,
        metadata={"help": "None则不需要提前计算 reference log p，路径不存在则需要提前计算，路径存在则直接导入数据"}
    )
    use_model_parallel: Optional[bool] = field(
        default=False,
        metadata={"help": "是否设置模型并行，当compute_on_8a100设置为True时，该选项失效"}
    )
    pre_compute_reference_columns: Optional[List[str]] = field(
        default_factory=lambda: ["reference_chosen_logps", "reference_rejected_logps",
                                 "reference_chosen_logps_per_token", "reference_rejected_logps_per_token",
                                 "reference_chosen_logits", "reference_rejected_logits",
                                 "reference_chosen_probs", "reference_rejected_probs",
                                 "reference_chosen_length", "reference_rejected_length"],
        metadata={"help": "预先计算的字段"},
    )
    pre_compute_reference_cut_end: Optional[str] = field(
        default=False,
        metadata={"help": "在计算reference时，是否截掉user-end部分"}
    )
    """
    接下来的参数控制是否在chosen上进行sft
    """
    sft_on_chosen: Optional[float] = field(
        default=None,
        metadata={"help": "在chosen上做sft占比损失的权重，小于1e-6则不计算"}
    )

    """
    接下来的参数主要用于标记assistant的回答区域（用于蒸馏）
    """
    user_start: Optional[str] = field(
        default="<|im_start|>system\n<|im_end|>\n<|im_start|>user\n",
        metadata={"help": "user prompt部分的模板开始标签，例如'<|im_start|>system\n<|im_end|>\n<|im_start|>user\n"}
    )
    user_end: Optional[str] = field(
        default="<|im_end|>\n",
        metadata={"help": "user prompt部分的模板开始标签，例如<|im_end|>\n"}
    )
    assistant_start: Optional[str] = field(
        default="<|im_start|>assistant\n",
        metadata={"help": "assistant回答部分的模板开始标签，例如<|im_start|>assistant\n"}
    )
    assistant_end: Optional[str] = field(
        default="<|im_end|>\n",
        metadata={"help": "assistant回答部分的模板结束标签，例如<|im_end|>\n"}
    )
    gamma: Optional[float] = field(
        default=1.6,
        metadata={"help": "gamma in Simpo"}
    )
    debug_mode: Optional[bool] = field(
        default=False,
    )
    eval_mode: Optional[bool] = field(
        default=False,
        metadata={"help": "set true when you wanna log all state"}
    )
    use_logits_or_probs: Optional[str] = field(
        default="logits",
    )
    fast_load: Optional[bool] = field(
        default=False,
    )
    evaluation_first: Optional[bool] = field(
        default=False,
    )


class PreCalRefDpoTrainer(DPOTrainer):
    def __dpo_trainer_init__(
            self,
            model: Optional[Union[PreTrainedModel, nn.Module, str]] = None,
            ref_model: Optional[Union[PreTrainedModel, nn.Module, str]] = None,
            beta: float = 0.1,
            label_smoothing: float = 0,
            loss_type: Literal["sigmoid", "hinge", "ipo", "kto_pair"] = "sigmoid",
            args: Optional[TrainingArguments] = None,
            data_collator: Optional[DataCollator] = None,
            label_pad_token_id: int = -100,
            padding_value: Optional[int] = None,
            truncation_mode: str = "keep_end",
            train_dataset: Optional[Dataset] = None,
            eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
            tokenizer: Optional[PreTrainedTokenizerBase] = None,
            model_init: Optional[Callable[[], PreTrainedModel]] = None,
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
            preprocess_logits_for_metrics: Optional[Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
            max_length: Optional[int] = None,
            max_prompt_length: Optional[int] = None,
            max_target_length: Optional[int] = None,
            peft_config: Optional[Dict] = None,
            is_encoder_decoder: Optional[bool] = None,
            disable_dropout: bool = True,
            generate_during_eval: bool = False,
            compute_metrics: Optional[Callable[[EvalLoopOutput], Dict]] = None,
            precompute_ref_log_probs: bool = False,
            dataset_num_proc: Optional[int] = None,
            model_init_kwargs: Optional[Dict] = None,
            ref_model_init_kwargs: Optional[Dict] = None,
            model_adapter_name: Optional[str] = None,
            ref_adapter_name: Optional[str] = None,
            reference_free: bool = False,
            force_use_ref_model: bool = False,
    ):
        self.is_vision_model = False
        self.args = args
        self.dataset_num_proc = dataset_num_proc
        self.is_encoder_decoder = False
        self.tokenizer = tokenizer
        self.label_pad_token_id = -100
        self.max_length = max_length
        # self.truncation_mode = "keep_end"
        self.max_prompt_length = max_prompt_length
        # with PartialState().local_main_process_first():
        #     # tokenize the dataset
        #     train_dataset = train_dataset.map(self.tokenize_row, num_proc=self.dataset_num_proc)
        #     if eval_dataset is not None:
        #         eval_dataset = eval_dataset.map(self.tokenize_row, num_proc=self.dataset_num_proc)

        if model_init_kwargs is None:
            model_init_kwargs = {}
        elif not isinstance(model, str):
            raise ValueError("You passed model_kwargs to the DPOTrainer. But your model is already instantiated.")

        if ref_model_init_kwargs is None:
            ref_model_init_kwargs = {}
        elif not isinstance(ref_model, str):
            raise ValueError(
                "You passed ref_model_kwargs to the DPOTrainer. But your ref_model is already instantiated."
            )

        if isinstance(model, str):
            warnings.warn(
                "You passed a model_id to the DPOTrainer. This will automatically create an "
                "`AutoModelForCausalLM` or a `PeftModel` (if you passed a `peft_config`) for you."
            )
            model = AutoModelForCausalLM.from_pretrained(model, **model_init_kwargs)

        if isinstance(ref_model, str):
            warnings.warn(
                "You passed a ref model_id to the DPOTrainer. This will automatically create an "
                "`AutoModelForCausalLM`"
            )
            ref_model = AutoModelForCausalLM.from_pretrained(ref_model, **ref_model_init_kwargs)

        # Initialize this variable to False. This helps tracking the case when `peft_module_casting_to_bf16`
        # has been called in order to properly call autocast if needed.
        self._peft_has_been_casted_to_bf16 = False

        if not is_peft_available() and peft_config is not None:
            raise ValueError(
                "PEFT is not installed and you passed a `peft_config` in the trainer's kwargs, please install it to use the PEFT models"
            )
        elif is_peft_available() and peft_config is not None:
            # if model is a peft model and we have a peft_config, we merge and unload it first
            if isinstance(model, PeftModel):
                model = model.merge_and_unload()

            if ref_model is not None and not force_use_ref_model:
                raise ValueError(
                    "You passed both a ref_model and a peft_config. For training PEFT adapters with DPO there is no need to pass a reference"
                    " model. Please pass `ref_model=None` in case you want to train PEFT adapters, or pass a ref_model with `force_use_ref_model=True` in DPOTrainer's init."
                    " if you want to use a different ref_model."
                )

            if getattr(model, "is_loaded_in_8bit", False) or getattr(model, "is_loaded_in_4bit", False):
                _support_gc_kwargs = hasattr(
                    args, "gradient_checkpointing_kwargs"
                ) and "gradient_checkpointing_kwargs" in list(
                    inspect.signature(prepare_model_for_kbit_training).parameters
                )

                prepare_model_kwargs = {"use_gradient_checkpointing": args.gradient_checkpointing}

                if _support_gc_kwargs:
                    prepare_model_kwargs["gradient_checkpointing_kwargs"] = args.gradient_checkpointing_kwargs

                model = prepare_model_for_kbit_training(model, **prepare_model_kwargs)
            elif getattr(args, "gradient_checkpointing", False):
                # For backward compatibility with older versions of transformers
                if hasattr(model, "enable_input_require_grads"):
                    model.enable_input_require_grads()
                else:

                    def make_inputs_require_grad(module, input, output):
                        output.requires_grad_(True)

                    model.get_input_embeddings().register_forward_hook(make_inputs_require_grad)

            # get peft model with the given config
            model = get_peft_model(model, peft_config)
            if args.bf16 and getattr(model, "is_loaded_in_4bit", False):
                peft_module_casting_to_bf16(model)
                # If args.bf16 we need to explicitly call `generate` with torch amp autocast context manager
                self._peft_has_been_casted_to_bf16 = True

        # For models that use gradient_checkpointing, we need to attach a hook that enables input
        # to explicitly have `requires_grad=True`, otherwise training will either silently
        # fail or completely fail.
        elif getattr(args, "gradient_checkpointing", False):
            # For backward compatibility with older versions of transformers
            if hasattr(model, "enable_input_require_grads"):
                model.enable_input_require_grads()
            else:

                def make_inputs_require_grad(module, input, output):
                    output.requires_grad_(True)

                model.get_input_embeddings().register_forward_hook(make_inputs_require_grad)

        if generate_during_eval and not is_wandb_available():
            raise ValueError(
                "`generate_during_eval=True` requires Weights and Biases to be installed."
                " Please install `wandb` to resolve."
            )

        if model is not None:
            self.is_encoder_decoder = model.config.is_encoder_decoder
        elif is_encoder_decoder is None:
            raise ValueError("When no model is provided, you need to pass the parameter is_encoder_decoder.")
        else:
            self.is_encoder_decoder = is_encoder_decoder

        self.is_peft_model = is_peft_available() and isinstance(model, PeftModel)
        self.model_adapter_name = model_adapter_name
        self.ref_adapter_name = ref_adapter_name
        self.reference_free = reference_free

        if ref_model:
            self.ref_model = ref_model
        elif self.is_peft_model or precompute_ref_log_probs or not self.args.compute_on_8a100:
            # The `model` with adapters turned off will be used as the reference model
            self.ref_model = None
        else:
            self.ref_model = None  # create_reference_model(model)

        if tokenizer is None:
            raise ValueError("tokenizer must be specified to tokenize a DPO dataset.")
        if max_length is None:
            warnings.warn(
                "`max_length` is not set in the DPOTrainer's init"
                " it will default to `512` by default, but you should do it yourself in the future.",
                UserWarning,
            )
            max_length = 512
        if max_prompt_length is None:
            warnings.warn(
                "`max_prompt_length` is not set in the DPOTrainer's init"
                " it will default to `128` by default, but you should do it yourself in the future.",
                UserWarning,
            )
            max_prompt_length = 128

        if max_target_length is None and self.is_encoder_decoder:
            warnings.warn(
                "When using an encoder decoder architecture, you should set `max_target_length` in the DPOTrainer's init"
                " it will default to `128` by default, but you should do it yourself in the future.",
                UserWarning,
            )
            max_target_length = 128

        if data_collator is None:
            data_collator = DPODataCollatorWithPadding(
                pad_token_id=tokenizer.pad_token_id,
                label_pad_token_id=label_pad_token_id,
                is_encoder_decoder=self.is_encoder_decoder,
            )

            if args.remove_unused_columns:
                args.remove_unused_columns = False
                # warn users
                warnings.warn(
                    "When using DPODataCollatorWithPadding, you should set `remove_unused_columns=False` in your TrainingArguments"
                    " we have set it for you, but you should do it yourself in the future.",
                    UserWarning,
                )

            self.use_dpo_data_collator = True
        else:
            self.use_dpo_data_collator = False

        if disable_dropout:
            disable_dropout_in_model(model)
            if self.ref_model is not None:
                disable_dropout_in_model(self.ref_model)

        self.max_length = max_length
        self.generate_during_eval = generate_during_eval
        self.label_pad_token_id = label_pad_token_id
        self.padding_value = padding_value if padding_value is not None else tokenizer.pad_token_id
        self.max_prompt_length = max_prompt_length
        self.truncation_mode = truncation_mode
        self.max_target_length = max_target_length
        self.tokenizer = tokenizer
        self.precompute_ref_log_probs = precompute_ref_log_probs

        # Since ref_logs are precomputed on the first call to get_train/eval_dataloader
        # keep track of first called to avoid computation of future calls
        self._precomputed_train_ref_log_probs = False
        self._precomputed_eval_ref_log_probs = False

        if loss_type in ["hinge", "ipo", "kto_pair"] and label_smoothing > 0:
            warnings.warn(
                "You are using a loss type that does not support label smoothing. Ignoring label_smoothing parameter."
            )

        self.beta = beta
        self.label_smoothing = label_smoothing
        self.loss_type = loss_type

        self._stored_metrics = defaultdict(lambda: defaultdict(list))

        # Compute that only on the main process for faster data processing.
        # see: https://github.com/huggingface/trl/pull/1255
        self.assistant_start_ids = self.tokenizer.encode(self.args.assistant_start, add_special_tokens=False)
        self.assistant_end_ids = self.tokenizer.encode(self.args.assistant_end, add_special_tokens=False)
        self.user_start_ids = self.tokenizer.encode(self.args.user_start, add_special_tokens=False)
        self.user_end_ids = self.tokenizer.encode(self.args.user_end, add_special_tokens=False)
        necessary_key = [
            "prompt", "chosen", "rejected",
            "chosen_input_ids", "chosen_attention_mask", "chosen_labels",
            "rejected_input_ids", "rejected_attention_mask", "rejected_labels",
            "prompt_input_ids", "prompt_attention_mask",
            "reference_chosen_logps", "reference_rejected_logps",
            "reference_chosen_probs", "reference_rejected_probs",
            "reference_chosen_logits", "reference_rejected_logits",
            "reference_chosen_length", "reference_rejected_length",
        ]
        if not (all([nk in train_dataset.column_names for nk in necessary_key]) and self.args.fast_load):
            with PartialState().local_main_process_first():
                # tokenize the dataset
                train_dataset = train_dataset.map(self.tokenize_row, num_proc=self.dataset_num_proc)
                if eval_dataset is not None:
                    eval_dataset = eval_dataset.map(self.tokenize_row, num_proc=self.dataset_num_proc)

        Trainer.__init__(
            self,
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
            optimizers=optimizers,
            preprocess_logits_for_metrics=preprocess_logits_for_metrics,
        )

        # Add tags for models that have been loaded with the correct transformers version
        if hasattr(self.model, "add_model_tags"):
            self.model.add_model_tags(self._tag_names)

        if not hasattr(self, "accelerator"):
            raise AttributeError(
                "Your `Trainer` does not have an `accelerator` object. Consider upgrading `transformers`."
            )

        # Deepspeed Zero-3 does not support precompute_ref_log_probs
        if self.is_deepspeed_enabled:
            if self.accelerator.state.deepspeed_plugin.zero_stage == 3 and self.precompute_ref_log_probs:
                raise ValueError(
                    "You cannot use `precompute_ref_log_probs=True` with Deepspeed ZeRO-3. Please set `precompute_ref_log_probs=False`."
                )

        else:
            if self.is_deepspeed_enabled:
                self.ref_model = self._prepare_deepspeed(self.ref_model)
            elif isinstance(self.ref_model, torch.nn.Module):
                self.ref_model = self.accelerator.prepare_model(self.ref_model, evaluation_mode=True)
            else:
                pass

    def __init__(self, **kwargs):
        self.pre_compute_period = kwargs.pop("pre_compute_period", False)
        self.__dpo_trainer_init__(**kwargs)  # Pass all other arguments using **kwargs
        label_smoothing = kwargs['args'].label_smoothing
        self.label_smoothing = label_smoothing
        # super().__init__(**kwargs)

    @staticmethod
    def get_per_token_logps(
            logits: torch.FloatTensor,
            labels: torch.LongTensor,
            label_pad_token_id: int = -100,
            is_encoder_decoder: bool = False,
    ) -> tuple[list[Tensor], Tensor, list[Tensor], Any]:
        """Compute the log probabilities of the given labels under the given logits.

        Args:
            logits: Logits of the model (unnormalized). Shape: (batch_size, sequence_length, vocab_size)
            labels: Labels for which to compute the log probabilities. Label tokens with a value of label_pad_token_id are ignored. Shape: (batch_size, sequence_length)
            average_log_prob: If True, return the average log probability per (non-masked) token. Otherwise, return the sum of the log probabilities of the (non-masked) tokens.
            label_pad_token_id: The label pad token id.
            is_encoder_decoder: Whether the model is an encoder-decoder model.

        Returns:
            A tensor of shape (batch_size,) containing the average/sum log probabilities of the given labels under the given logits.
        """
        if logits.shape[:-1] != labels.shape:
            raise ValueError("Logits (batch and sequence length dim) and labels must have the same shape.")

        if not is_encoder_decoder:
            labels = labels[:, 1:].clone()  #
            logits = logits[:, :-1, :]
        loss_mask = labels != label_pad_token_id

        # dummy token; we'll ignore the losses on these tokens later
        labels[labels == label_pad_token_id] = 0
        per_token_logps = torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)
        per_token_logps_ = []
        logits_ = []
        for i in range(len(loss_mask)):
            per_token_logps_.append(per_token_logps[i][loss_mask[i]])
            logits_.append(logits[i][loss_mask[i]])
            # per_token_logps_.append([float(per_token_logps[i][j]) for j in range(len(loss_mask[i])) if loss_mask[i][j]])

        return per_token_logps_, (per_token_logps * loss_mask).sum(-1), logits_, loss_mask.sum(-1)

    def tokenize_row(self, feature, model: Optional[Union[PreTrainedModel, nn.Module]] = None) -> Dict:
        """Tokenize a single row from a DPO specific dataset.

        At this stage, we don't convert to PyTorch tensors yet; we just handle the truncation
        in case the prompt + chosen or prompt + rejected responses is/are too long. First
            we truncate the prompt; if we're still too long, we truncate the chosen/rejected.

        We also create the labels for the chosen/rejected responses, which are of length equal to
            the sum of the length of the prompt and the chosen/rejected response, with
            label_pad_token_id  for the prompt tokens.
        """
        necessary_key = [
            "prompt", "chosen", "rejected",
            "chosen_input_ids", "chosen_attention_mask", "chosen_labels",
            "rejected_input_ids", "rejected_attention_mask", "rejected_labels",
            "prompt_input_ids", "prompt_attention_mask",
            "reference_chosen_logps", "reference_rejected_logps",
            "reference_chosen_probs", "reference_rejected_probs",
            "reference_chosen_logits", "reference_rejected_logits",
            "reference_chosen_length", "reference_rejected_length",
        ]
        # if all([nk in feature.keys() for nk in necessary_key]) and self.args.fast_load:
        if self.args.fast_load:
            return feature
        batch = {}
        prompt = feature["prompt"]
        chosen = feature["chosen"]
        rejected = feature["rejected"]
        # prompt_raw = prompt[len(self.args.user_start):] if prompt.startswith(self.args.user_start) else prompt
        # prompt_raw = prompt_raw[: -len(self.args.user_end)] if prompt.endswith(self.args.user_end) else prompt_raw
        # chosen_raw = chosen[len(self.args.assistant_start):] if chosen.startswith(self.args.assistant_start) else chosen
        # chosen_raw = chosen_raw[: -len(self.args.assistant_end)] if chosen.endswith(
        #     self.args.assistant_end) else chosen_raw
        # rejected_raw = rejected[len(self.args.assistant_start):] if rejected.startswith(
        #     self.args.assistant_start) else chosen
        # rejected_raw = rejected_raw[: -len(self.args.assistant_end)] if rejected.endswith(
        #     self.args.assistant_end) else rejected_raw
        #
        # prompt = self.args.user_start + prompt_raw + self.args.user_end
        # chosen = self.args.assistant_start + chosen_raw + self.args.assistant_end
        # rejected = self.args.assistant_start + rejected_raw + self.args.assistant_end

        if "reference_chosen_logps" in feature and "reference_rejected_logps" in feature:
            batch["reference_chosen_logps"] = feature["reference_chosen_logps"]
            batch["reference_rejected_logps"] = feature["reference_rejected_logps"]

        if "reference_chosen_logits" in feature and "reference_rejected_logits" in feature:
            batch["reference_chosen_logits"] = feature["reference_chosen_logits"]
            batch["reference_rejected_logits"] = feature["reference_rejected_logits"]

        if "reference_chosen_probs" in feature and "reference_rejected_probs" in feature:
            batch["reference_chosen_probs"] = feature["reference_chosen_probs"]
            batch["reference_rejected_probs"] = feature["reference_rejected_probs"]

        if "reference_chosen_length" in feature and "reference_rejected_length" in feature:
            batch["reference_chosen_length"] = feature["reference_chosen_length"]
            batch["reference_rejected_length"] = feature["reference_rejected_length"]

        if not self.is_encoder_decoder:
            # Check issues below for more details
            #  1. https://github.com/huggingface/trl/issues/907
            #  2. https://github.com/EleutherAI/lm-evaluation-harness/pull/531#issuecomment-1595586257
            #  3. https://github.com/LianjiaTech/BELLE/issues/337

            if not isinstance(prompt, str):
                raise ValueError(f"prompt should be an str but got {type(prompt)}")
            prompt_tokens = self.tokenizer(prompt, add_special_tokens=False)
            prompt_tokens = {f"prompt_{k}": v for k, v in prompt_tokens.items()}

            if not isinstance(chosen, str):
                raise ValueError(f"chosen should be an str but got {type(chosen)}")
            chosen_tokens = self.build_tokenized_answer(prompt, chosen)

            if not isinstance(rejected, str):
                raise ValueError(f"rejected should be an str but got {type(rejected)}")
            rejected_tokens = self.build_tokenized_answer(prompt, rejected)

            # Last prompt token might get merged by tokenizer and
            # it should not be included for generation if that happens
            prompt_len_input_ids = len(prompt_tokens["prompt_input_ids"])

            chosen_prompt_len_input_ids = len(chosen_tokens["prompt_input_ids"])
            rejected_prompt_len_input_ids = len(rejected_tokens["prompt_input_ids"])
            prompt_len_input_ids = min(chosen_prompt_len_input_ids, rejected_prompt_len_input_ids)

            for k, v in prompt_tokens.items():
                prompt_tokens[k] = v[:prompt_len_input_ids]

            # Make sure prompts only have one different token at most an
            # and length only differs by 1 at most
            num_diff_tokens = sum(
                [a != b for a, b in zip(chosen_tokens["prompt_input_ids"], rejected_tokens["prompt_input_ids"])]
            )
            num_diff_len = abs(chosen_prompt_len_input_ids - rejected_prompt_len_input_ids)
            if num_diff_tokens > 1 or num_diff_len > 1:
                raise ValueError(
                    "Chosen and rejected prompt_input_ids might only differ on the "
                    "last token due to tokenizer merge ops."
                )

            # add BOS token to head of prompt
            if prompt_tokens["prompt_input_ids"][0] != self.tokenizer.bos_token_id:
                prompt_tokens["prompt_input_ids"] = [self.tokenizer.bos_token_id] + prompt_tokens["prompt_input_ids"]
                prompt_tokens["prompt_attention_mask"] = [1] + prompt_tokens["prompt_attention_mask"]
            if chosen_tokens["prompt_input_ids"][0] != self.tokenizer.bos_token_id:
                chosen_tokens["prompt_input_ids"] = [self.tokenizer.bos_token_id] + chosen_tokens["prompt_input_ids"]
                chosen_tokens["prompt_attention_mask"] = [1] + chosen_tokens["prompt_attention_mask"]
            if rejected_tokens["prompt_input_ids"][0] != self.tokenizer.bos_token_id:
                rejected_tokens["prompt_input_ids"] = [self.tokenizer.bos_token_id] + rejected_tokens[
                    "prompt_input_ids"]
                rejected_tokens["prompt_attention_mask"] = [1] + rejected_tokens["prompt_attention_mask"]

            # add EOS token to end of answer
            if self.args.pre_compute_reference_cut_end and self.pre_compute_period:
                pass
            else:
                # if chosen_tokens["input_ids"][-1] != self.tokenizer.eos_token_id:
                #     chosen_tokens["input_ids"].append(self.tokenizer.eos_token_id)
                #     chosen_tokens["attention_mask"].append(1)
                #
                # if rejected_tokens["input_ids"][-1] != self.tokenizer.eos_token_id:
                #     rejected_tokens["input_ids"].append(self.tokenizer.eos_token_id)
                #     rejected_tokens["attention_mask"].append(1)

                if self.tokenizer.eos_token_id not in chosen_tokens["input_ids"][-5:]:
                    chosen_tokens["input_ids"].append(self.tokenizer.eos_token_id)
                    chosen_tokens["attention_mask"].append(1)

                if self.tokenizer.eos_token_id not in rejected_tokens["input_ids"][-5:]:
                    rejected_tokens["input_ids"].append(self.tokenizer.eos_token_id)
                    rejected_tokens["attention_mask"].append(1)
            longer_response_length = max(len(chosen_tokens["input_ids"]), len(rejected_tokens["input_ids"]))

            # if combined sequence is too long, truncate the prompt
            for answer_tokens in [chosen_tokens, rejected_tokens, prompt_tokens]:
                if len(answer_tokens["prompt_input_ids"]) + longer_response_length > self.max_length:
                    if self.truncation_mode == "keep_start":
                        for k in ["prompt_input_ids", "prompt_attention_mask"]:
                            answer_tokens[k] = answer_tokens[k][: self.max_prompt_length]
                    elif self.truncation_mode == "keep_end":
                        for k in ["prompt_input_ids", "prompt_attention_mask"]:
                            answer_tokens[k] = answer_tokens[k][-self.max_prompt_length:]
                    else:
                        raise ValueError(f"Unknown truncation mode: {self.truncation_mode}")

            # if that's still too long, truncate the response
            for answer_tokens in [chosen_tokens, rejected_tokens]:
                if len(answer_tokens["prompt_input_ids"]) + longer_response_length > self.max_length:
                    for k in ["input_ids", "attention_mask"]:
                        answer_tokens[k] = answer_tokens[k][: self.max_length - self.max_prompt_length]

            # Create labels
            chosen_sequence_tokens = {
                k: chosen_tokens[f"prompt_{k}"] + chosen_tokens[k] for k in ["input_ids", "attention_mask"]
            }
            rejected_sequence_tokens = {
                k: rejected_tokens[f"prompt_{k}"] + rejected_tokens[k] for k in ["input_ids", "attention_mask"]
            }
            chosen_sequence_tokens["labels"] = chosen_sequence_tokens["input_ids"][:]
            chosen_sequence_tokens["labels"][: len(chosen_tokens["prompt_input_ids"])] = [
                                                                                             self.label_pad_token_id
                                                                                         ] * len(
                chosen_tokens["prompt_input_ids"])
            rejected_sequence_tokens["labels"] = rejected_sequence_tokens["input_ids"][:]
            rejected_sequence_tokens["labels"][: len(rejected_tokens["prompt_input_ids"])] = [
                                                                                                 self.label_pad_token_id
                                                                                             ] * len(
                rejected_tokens["prompt_input_ids"])

            for k, toks in {
                "chosen_": chosen_sequence_tokens,
                "rejected_": rejected_sequence_tokens,
                "": prompt_tokens,
            }.items():
                for type_key, tokens in toks.items():
                    if type_key == "token_type_ids":
                        continue
                    batch[f"{k}{type_key}"] = tokens

        else:
            chosen_tokens = self.tokenizer(
                chosen, truncation=True, max_length=self.max_target_length, add_special_tokens=True
            )
            rejected_tokens = self.tokenizer(
                rejected, truncation=True, max_length=self.max_target_length, add_special_tokens=True
            )
            prompt_tokens = self.tokenizer(
                prompt, truncation=True, max_length=self.max_prompt_length, add_special_tokens=True
            )

            batch["chosen_labels"] = chosen_tokens["input_ids"]
            batch["rejected_labels"] = rejected_tokens["input_ids"]
            batch["prompt_input_ids"] = prompt_tokens["input_ids"]
            batch["prompt_attention_mask"] = prompt_tokens["attention_mask"]

            if model is not None and hasattr(model, "prepare_decoder_input_ids_from_labels"):
                batch["rejected_decoder_input_ids"] = model.prepare_decoder_input_ids_from_labels(
                    labels=torch.tensor(batch["rejected_labels"])
                )
                batch["chosen_decoder_input_ids"] = model.prepare_decoder_input_ids_from_labels(
                    labels=torch.tensor(batch["chosen_labels"])
                )

        return batch

    def concatenated_forward_custom(
            self, model: nn.Module, batch: Dict[str, Union[List, torch.LongTensor]]
    ) -> Tuple[Tensor, Tensor, List[Tensor], List[Tensor], Tensor, Tensor]:
        """Run the given model on the given batch of inputs, concatenating the chosen and rejected inputs together.

        We do this to avoid doing two forward passes, because it's faster for FSDP.
        """
        concatenated_batch = self.concatenated_inputs(
            batch,
            is_encoder_decoder=self.is_encoder_decoder,
            label_pad_token_id=self.label_pad_token_id,
            padding_value=self.padding_value,
            device=self.accelerator.device,
        )
        len_chosen = batch["chosen_labels"].shape[0]

        model_kwargs = (
            {
                "labels": concatenated_batch["concatenated_labels"],
                "decoder_input_ids": concatenated_batch.pop("concatenated_decoder_input_ids", None),
            }
            if self.is_encoder_decoder
            else {}
        )
        all_logits = model(
            concatenated_batch["concatenated_input_ids"],
            attention_mask=concatenated_batch["concatenated_attention_mask"],
            use_cache=False,
            **model_kwargs,
        ).logits

        all_logps_per_token, all_logps, all_logits, all_lengths = self.get_per_token_logps(
            all_logits,
            concatenated_batch["concatenated_labels"],
            is_encoder_decoder=self.is_encoder_decoder,
            label_pad_token_id=self.label_pad_token_id,
        )

        chosen_logps = all_logps[:len_chosen]
        rejected_logps = all_logps[len_chosen:]

        chosen_logps_per_token = all_logps_per_token[:len_chosen]
        rejected_logps_per_token = all_logps_per_token[len_chosen:]

        chosen_logits = all_logits[:len_chosen]
        rejected_logits = all_logits[len_chosen:]

        chosen_length = all_lengths[:len_chosen]
        rejected_length = all_lengths[len_chosen:]

        return chosen_logps, rejected_logps, chosen_logps_per_token, rejected_logps_per_token, chosen_logits, rejected_logits, chosen_length, rejected_length

    def compute_reference_custom(self, model, padded_batch: Dict) -> Dict[str, List]:
        """
        Computes :
            reference_chosen_logps, reference_rejected_logps,
            reference_chosen_logps_per_token, reference_rejected_logps_per_token",
            reference_chosen_logits, reference_rejected_logits
        of the reference model for a single padded batch of a DPO specific dataset.
        """
        chosen_logps, rejected_logps, \
            chosen_logps_per_token, rejected_logps_per_token, \
            chosen_logits, rejected_logits, chosen_length, rejected_length = self.concatenated_forward_custom(
            model, padded_batch)

        for i in range(len(chosen_logps_per_token)):
            if abs(sum(rejected_logps_per_token[i]) - rejected_logps[i]) > 1e2:
                print(
                    f"Batch {i}: Sum of per-token rejected logps = {sum(rejected_logps_per_token[i])}, Rejected logps = {rejected_logps[i]}")
            if abs(sum(chosen_logps_per_token[i]) - chosen_logps[i]) > 1e-2:
                print(
                    f"Batch {i}: Sum of per-token chosen logps = {sum(chosen_logps_per_token[i])}, Chosen logps = {chosen_logps[i]}")

        return {
            "reference_chosen_logps": chosen_logps.detach().cpu().numpy().tolist(),
            "reference_rejected_logps": rejected_logps.detach().cpu().numpy().tolist(),
            "reference_chosen_logps_per_token": [clpt.detach().cpu().numpy().tolist() for clpt in
                                                 chosen_logps_per_token],
            "reference_rejected_logps_per_token": [rlpt.detach().cpu().numpy().tolist() for rlpt in
                                                   rejected_logps_per_token],
            "reference_chosen_logits": chosen_logits,
            "reference_rejected_logits": rejected_logits,
            "reference_chosen_length": chosen_length.cpu().numpy().tolist(),
            "reference_rejected_length": rejected_length.cpu().numpy().tolist()}

    def pre_compute_and_save_reference(self, dataset: datasets.Dataset, pre_compute_reference_columns: List[str]):
        """
        Compute reference log probabilities for the given dataset and save them in the dataset.
        """

        dataloader_params = {
            "batch_size": self.args.per_device_eval_batch_size,
            "collate_fn": self.data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "shuffle": False,
        }

        # # 去除最后的 user_end_ids token
        # if self.args.pre_compute_reference_cut_end:
        #     def cut_end(row):
        #         rst = {}
        #         if row['chosen_input_ids'].endswith(self.user_end_ids):
        #             rst['chosen_input_ids'] = row['chosen_input_ids'][: -len(self.user_end_ids)]
        #             rst['chosen_attention_mask'] = row['chosen_attention_mask'][: -len(self.user_end_ids)]
        #             rst['chosen_labels'] = row['chosen_labels'][: -len(self.user_end_ids)]
        #         else:
        #             rst['chosen_input_ids'] = row['chosen_input_ids']
        #             rst['chosen_attention_mask'] = row['chosen_attention_mask']
        #             rst['chosen_labels'] = row['chosen_labels']
        #
        #         if row['rejected_input_ids'].endswith(self.user_end_ids):
        #             rst['rejected_input_ids'] = row['rejected_input_ids'][: -len(self.user_end_ids)]
        #             rst['rejected_attention_mask'] = row['rejected_attention_mask'][: -len(self.user_end_ids)]
        #             rst['rejected_labels'] = row['rejected_labels'][: -len(self.user_end_ids)]
        #         else:
        #             rst['rejected_input_ids'] = row['rejected_input_ids']
        #             rst['rejected_attention_mask'] = row['rejected_attention_mask']
        #             rst['rejected_labels'] = row['rejected_labels']
        #         return rst
        #     dataset = dataset.map(cut_end)

        # Prepare DataLoader
        data_loader = self.accelerator.prepare(DataLoader(dataset, **dataloader_params))

        # calculate the reference
        pre_compute_reference_info: Dict[str, List] = {k: [] for k in pre_compute_reference_columns}
        for padded_batch in tqdm(data_loader, desc=f"Computing {', '.join(pre_compute_reference_columns)}"):
            with torch.no_grad():
                model_batch_info = self.compute_reference_custom(self.model, padded_batch)
                model_batch_info['reference_chosen_logps'], \
                    model_batch_info['reference_rejected_logps'], \
                    model_batch_info['reference_chosen_logps_per_token'], \
                    model_batch_info['reference_rejected_logps_per_token'], \
                    model_batch_info['reference_chosen_logits'], \
                    model_batch_info['reference_rejected_logits'], \
                    model_batch_info['reference_chosen_length'], \
                    model_batch_info['reference_rejected_length'] = \
                    self.accelerator.gather_for_metrics((
                        model_batch_info['reference_chosen_logps'],
                        model_batch_info['reference_rejected_logps'],
                        model_batch_info['reference_chosen_logps_per_token'],
                        model_batch_info['reference_rejected_logps_per_token'],
                        model_batch_info['reference_chosen_logits'],
                        model_batch_info['reference_rejected_logits'],
                        model_batch_info['reference_chosen_length'],
                        model_batch_info['reference_rejected_length']))
                if self.ref_model is not None:
                    reference_batch_info = self.compute_reference_custom(self.ref_model, padded_batch)
                    reference_batch_info['reference_chosen_logps'], \
                        reference_batch_info['reference_rejected_logps'], \
                        reference_batch_info['reference_chosen_logps_per_token'], \
                        reference_batch_info['reference_rejected_logps_per_token'], \
                        reference_batch_info['reference_chosen_logits'], \
                        reference_batch_info['reference_rejected_logits'], \
                        reference_batch_info['reference_chosen_length'], \
                        reference_batch_info['reference_rejected_length'] = \
                        self.accelerator.gather_for_metrics((
                            reference_batch_info['reference_chosen_logps'],
                            reference_batch_info['reference_rejected_logps'],
                            reference_batch_info['reference_chosen_logps_per_token'],
                            reference_batch_info['reference_rejected_logps_per_token'],
                            reference_batch_info['reference_chosen_logits'],
                            reference_batch_info['reference_rejected_logits'],
                            reference_batch_info['reference_chosen_length'],
                            reference_batch_info['reference_rejected_length']))

                # tmp_chosen = model_batch_info['reference_chosen_logits']
                # tmp_rejected = model_batch_info['reference_rejected_logits']
                # debug: [rcp['values'][0] for rcp in model_batch_info['reference_chosen_probs'][0]], [math.exp(rclpt) for rclpt in model_batch_info['reference_chosen_logps_per_token'][0]]
                # 对 reference 的 logits 进行压缩，节省空间
                model_batch_info['reference_chosen_probs'] = [compress_probs(rcl) for rcl in
                                                                  model_batch_info['reference_chosen_logits']]
                model_batch_info['reference_rejected_probs'] = [compress_probs(rrl) for rrl in
                                                                    model_batch_info['reference_rejected_logits']]
                model_batch_info['reference_chosen_logits'] = [compress_logits(rcl) for rcl in
                                                                   model_batch_info['reference_chosen_logits']]
                model_batch_info['reference_rejected_logits'] = [compress_logits(rrl) for rrl in
                                                                     model_batch_info['reference_rejected_logits']]

                # for i, (rcl, rrl, cl, rl) in enumerate(zip(
                #     model_batch_info['reference_chosen_logits'],
                #     model_batch_info['reference_rejected_logits'],
                #     padded_batch["chosen_labels"],
                #     padded_batch["rejected_labels"]
                # )):
                #     assert (softmax(tmp_chosen[i].cpu()) - softmax(restore_logits(rcl, seq_len=len(rcl), vocab_size=len(self.tokenizer)))).abs().sum() < 1
                #     assert (softmax(tmp_rejected[i].cpu()) - softmax(restore_logits(rrl, seq_len=len(rrl), vocab_size=len(self.tokenizer)))).abs().sum() < 1
                #     assert len(rcl) == (cl != -100).sum() and len(rrl) == (rl != -100).sum()
                #     cnt = 0
                #     for rcli, cli in zip(rcl, cl[cl != -100]):
                #         cnt += int(cli) in rcli['indices']
                #     print("\tchosen: ", round(float(cnt / (cl != -100).sum()) * 100, 2), "%", end="\t")
                #     cnt = 0
                #     for rrli, rli in zip(rrl, rl[rl != -100]):
                #         cnt += int(rli) in rrli['indices']
                #     print("rejected", round(float(cnt / (rl != -100).sum()) * 100, 2), "%")
            for column in pre_compute_reference_columns:
                pre_compute_reference_info[column].append(model_batch_info[column])

        for column in pre_compute_reference_columns:
            dataset = dataset.add_column(name=column, column=pre_compute_reference_info[column])

        return dataset

    @staticmethod
    def get_batch_logps(
            logits: torch.FloatTensor,
            labels: torch.LongTensor,
            average_log_prob: bool = False,
            label_pad_token_id: int = -100,
            is_encoder_decoder: bool = False,
    ) -> Tensor | Any:
        """
        Compute the log probabilities of the given labels under the given logits.

        Args:
            logits: Logits of the model (unnormalized). Shape: (batch_size, sequence_length, vocab_size)
            labels: Labels for which to compute the log probabilities. Label tokens with a value of label_pad_token_id are ignored. Shape: (batch_size, sequence_length)
            average_log_prob: If True, return the average log probability per (non-masked) token. Otherwise, return the sum of the log probabilities of the (non-masked) tokens.
            label_pad_token_id: The label pad token id.
            is_encoder_decoder: Whether the model is an encoder-decoder model.

        Returns:
            A tensor of shape (batch_size,) containing the average/sum log probabilities of the given labels under the given logits.
        """
        if logits.shape[:-1] != labels.shape:
            raise ValueError("Logits (batch and sequence length dim) and labels must have the same shape.")

        if not is_encoder_decoder:
            labels = labels[:, 1:].clone()
            logits = logits[:, :-1, :]
        loss_mask = labels != label_pad_token_id

        # dummy token; we'll ignore the losses on these tokens later
        labels[labels == label_pad_token_id] = 0

        per_token_logps = torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)

        if average_log_prob:
            return (per_token_logps * loss_mask).sum(-1) / loss_mask.sum(-1)
        else:
            return (per_token_logps * loss_mask).sum(-1), loss_mask.sum(-1)

    def sft_loss(self, chosen_logits: "torch.FloatTensor", chosen_labels: "torch.LongTensor") -> "torch.Tensor":
        r"""
        Computes supervised cross-entropy loss of given labels under the given logits.

        Returns:
            A tensor of shape (batch_size,) containing the cross-entropy loss of each samples.
        """
        all_logps, _ = self.get_batch_logps(chosen_logits, chosen_labels, average_log_prob=True)
        return -all_logps

    def get_batch_loss_metrics(
            self,
            model,
            batch: Dict[str, Union[List, torch.LongTensor]],
            train_eval: Literal["train", "eval.yaml"] = "train",
    ):
        """Compute the DPO loss and other metrics for the given batch of inputs for train or test."""
        prefix = "eval_" if train_eval == "eval.yaml" else ""
        metrics = {}

        (
            policy_chosen_logps,
            policy_rejected_logps,
            policy_chosen_logits,
            policy_rejected_logits, chosen_logps_avg
        ) = self.concatenated_forward(model, batch)

        # if reference_chosen_logps and reference_rejected_logps in batch use them, otherwise use the reference model
        if "reference_chosen_logps" in batch and "reference_rejected_logps" in batch:
            reference_chosen_logps = batch["reference_chosen_logps"]
            reference_rejected_logps = batch["reference_rejected_logps"]
        else:
            with torch.no_grad():
                if self.ref_model is None:
                    with self.null_ref_context():
                        (
                            reference_chosen_logps,
                            reference_rejected_logps,
                            _,
                            _,
                        ) = self.concatenated_forward(self.model, batch)
                else:
                    (
                        reference_chosen_logps,
                        reference_rejected_logps,
                        _,
                        _,
                    ) = self.concatenated_forward(self.ref_model, batch)

        losses, chosen_rewards, rejected_rewards = self.dpo_loss(
            policy_chosen_logps,
            policy_rejected_logps,
            reference_chosen_logps,
            reference_rejected_logps,
        )
        if self.args.sft_on_chosen > 1e-6:
            batch_size = batch["input_ids"].size(0) // 2
            chosen_labels, _ = batch["labels"].split(batch_size, dim=0)
            metrics[f"{prefix}loss/dpo"] = losses.mean().cpu()

            sft_loss = self.sft_loss(policy_chosen_logits, chosen_labels)
            losses += self.args.sft_on_chosen * sft_loss
            metrics[f"{prefix}loss/sft"] = sft_loss.mean().cpu()

        reward_accuracies = (chosen_rewards > rejected_rewards).float()
        metrics[f"{prefix}rewards/chosen"] = chosen_rewards.mean().cpu()
        metrics[f"{prefix}rewards/rejected"] = rejected_rewards.mean().cpu()
        metrics[f"{prefix}rewards/accuracies"] = reward_accuracies.mean().cpu()
        metrics[f"{prefix}rewards/margins"] = (chosen_rewards - rejected_rewards).mean().cpu()
        metrics[f"{prefix}logps/rejected"] = policy_rejected_logps.detach().mean().cpu()
        metrics[f"{prefix}logps/chosen"] = policy_chosen_logps.detach().mean().cpu()
        metrics[f"{prefix}logits/rejected"] = policy_rejected_logits.detach().mean().cpu()
        metrics[f"{prefix}logits/chosen"] = policy_chosen_logits.detach().mean().cpu()

        return losses.mean(), metrics


def direct_training(dpo_trainer_type, model_args, data_args, training_args, device_map):
    """不提前计算 reference，直接进行 DPO 训练"""
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    log_level = training_args.get_process_log_level()
    logger.setLevel(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()

    # Log on each process the small summary:
    logger.info(f"Model parameters {model_args}")
    logger.info(f"Data parameters {data_args}")
    logger.info(f"Training/evaluation parameters {training_args}")

    # Check for last checkpoint
    last_checkpoint = get_checkpoint(training_args)
    if last_checkpoint is not None and training_args.resume_from_checkpoint is None:
        logger.info(f"Checkpoint detected, resuming training at {last_checkpoint=}.")

    # Set seed for reproducibility
    set_seed(training_args.seed)

    ###############
    # Load datasets
    ###############
    try:
        raw_datasets = datasets.DatasetDict.load_from_disk(data_args.tokenized_dataset)
        data_args.truncation_side = "left"  # Truncate from left to ensure we don't lose labels in final turn
        tokenizer = get_tokenizer(model_args, data_args)
    except (TypeError, ValueError, FileNotFoundError):
        raw_datasets = get_datasets(
            data_args, shuffle=False,
            splits=data_args.dataset_splits,
            configs=data_args.dataset_configs,
            columns_to_keep=["messages", "chosen", "rejected", "prompt", "completion", "label"],
        )
        logger.info(
            f"Training on the following splits: {[split + ' : ' + str(dset.num_rows) for split, dset in raw_datasets.items()]}"
        )
        column_names = list(raw_datasets["train"].features)

        #####################################
        # Load tokenizer and process datasets
        #####################################
        data_args.truncation_side = "left"  # Truncate from left to ensure we don't lose labels in final turn
        tokenizer = get_tokenizer(model_args, data_args)

        #####################
        # Apply chat template
        #####################
        raw_datasets = raw_datasets.map(
            apply_chat_template,
            fn_kwargs={
                "tokenizer": tokenizer,
                "task": "dpo",
                "auto_insert_empty_system_msg": data_args.auto_insert_empty_system_msg,
            },
            num_proc=data_args.preprocessing_num_workers,
            remove_columns=column_names,
            desc="Formatting comparisons with prompt template",
        )

        ##########################
        # Decontaminate benchmarks
        ##########################
        num_raw_train_samples = len(raw_datasets["train"])
        raw_datasets = raw_datasets.filter(
            # decontaminate_humaneval,
            fn_kwargs={"text_column": "text_chosen"},
            batched=True,
            batch_size=10_000,
            num_proc=1,
            desc="Decontaminating HumanEval samples",
        )
        num_filtered_train_samples = num_raw_train_samples - len(raw_datasets["train"])
        logger.info(
            f"Decontaminated {num_filtered_train_samples} ({num_filtered_train_samples / num_raw_train_samples * 100:.2f}%) samples from the training set."
        )

        # Replace column names with what TRL needs, text_chosen -> chosen and text_rejected -> rejected
        for split in ["train", "test"]:
            raw_datasets[split] = raw_datasets[split].rename_columns(
                {"text_prompt": "prompt", "text_chosen": "chosen", "text_rejected": "rejected"}
            )

        if data_args.tokenized_dataset is not None:
            raw_datasets.save_to_disk(data_args.tokenized_dataset)

    # Log a few random samples from the training set:
    for index in random.sample(range(len(raw_datasets["train"])), 3):
        logger.info(f"Prompt sample {index} of the raw training set:\n\n{raw_datasets['train'][index]['prompt']}")
        logger.info(f"Chosen sample {index} of the raw training set:\n\n{raw_datasets['train'][index]['chosen']}")
        logger.info(f"Rejected sample {index} of the raw training set:\n\n{raw_datasets['train'][index]['rejected']}")

    torch_dtype = (
        model_args.torch_dtype if model_args.torch_dtype in ["auto", None] else getattr(torch, model_args.torch_dtype)
    )
    quantization_config = get_quantization_config(model_args)

    model_kwargs = dict(
        revision=model_args.model_revision,
        trust_remote_code=model_args.trust_remote_code,
        # use_flash_attention_2=model_args.use_flash_attention_2,
        torch_dtype=torch_dtype,
        use_cache=False if training_args.gradient_checkpointing else True,
        device_map=device_map,
        quantization_config=quantization_config,
    )

    model = model_args.model_name_or_path
    if is_adapter_model(model, model_args.model_revision) is True:
        logger.info(f"Loading SFT adapter for {model_args.model_name_or_path=}")
        peft_config = PeftConfig.from_pretrained(model_args.model_name_or_path, revision=model_args.model_revision)
        model_kwargs = dict(
            revision=model_args.base_model_revision,
            trust_remote_code=model_args.trust_remote_code,
            # use_flash_attention_2=model_args.use_flash_attention_2,
            torch_dtype=torch_dtype,
            use_cache=False if training_args.gradient_checkpointing else True,
            device_map=device_map,
            quantization_config=quantization_config,
        )
        base_model = AutoModelForCausalLM.from_pretrained(
            peft_config.base_model_name_or_path,
            **model_kwargs,
        )
        model = PeftModel.from_pretrained(
            base_model,
            model_args.model_name_or_path,
            revision=model_args.model_revision,
        )
        model_kwargs = None

    ref_model = model
    ref_model_kwargs = model_kwargs

    if model_args.use_peft is True:
        ref_model = None
        ref_model_kwargs = None

    #########################
    # Instantiate DPO trainer
    #########################
    if training_args.debug_mode:
        raw_datasets["train"] = raw_datasets["train"].select(range(100))
        raw_datasets["test"] = raw_datasets["test"].select(range(100))

    trainer = dpo_trainer_type(
        model,
        ref_model,
        model_init_kwargs=model_kwargs,
        ref_model_init_kwargs=ref_model_kwargs,
        args=training_args,
        beta=training_args.beta,
        train_dataset=raw_datasets["train"],
        eval_dataset=raw_datasets["test"],
        tokenizer=tokenizer,
        max_length=training_args.max_length,
        max_prompt_length=training_args.max_prompt_length,
        precompute_ref_log_probs=True,
        peft_config=get_peft_config(model_args),
        loss_type=training_args.loss_type,
    )

    ###############
    # Training loop
    ###############
    if training_args.num_train_epochs > 0 or training_args.max_steps > 0:
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif last_checkpoint is not None:
            checkpoint = last_checkpoint
        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        metrics = train_result.metrics
        metrics["train_samples"] = len(raw_datasets["train"])
        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

        logger.info("*** Training complete ***")

        ##################################
        # Save model and create model card
        ##################################
        logger.info("*** Save model ***")
        trainer.save_model(training_args.output_dir)
        # trainer.model.save_pretrained(training_args.output_dir, is_main_process=True)
        logger.info(f"Model saved to {training_args.output_dir}")

        # Save everything else on main process
        kwargs = {
            "finetuned_from": model_args.model_name_or_path,
            "dataset": list(data_args.dataset_mixer.keys()),
            "dataset_tags": list(data_args.dataset_mixer.keys()),
            "tags": ["alignment-handbook"],
        }
        if trainer.accelerator.is_main_process:
            trainer.create_model_card(**kwargs)
            # Restore k,v cache for fast inference
            trainer.model.config.use_cache = True
            trainer.model.config.save_pretrained(training_args.output_dir)

    ##########
    # Evaluate
    ##########
    if training_args.do_eval:
        logger.info("*** Evaluate ***")
        metrics = trainer.evaluate()
        metrics["eval_samples"] = len(raw_datasets["test"])
        trainer.log_metrics("eval.yaml", metrics)
        trainer.save_metrics("eval.yaml", metrics)

    if training_args.push_to_hub is True:
        logger.info("Pushing to hub...")
        trainer.push_to_hub(**kwargs)

    logger.info("*** Training complete! ***")


def pre_calculate_ref_training(dpo_trainer_type, model_args, data_args, training_args, device_map):
    """提前计算 reference 的 DPO 训练"""
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    log_level = training_args.get_process_log_level()
    logger.setLevel(log_level)
    transformers.utils.logging.set_verbosity(log_level)
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()
    logger.info(f"Model parameters {model_args}")
    logger.info(f"Data parameters {data_args}")
    logger.info(f"Training/evaluation parameters {training_args}")
    last_checkpoint = get_checkpoint(training_args)
    if last_checkpoint is not None and training_args.resume_from_checkpoint is None:
        logger.info(f"Checkpoint detected, resuming training at {last_checkpoint=}.")
    set_seed(training_args.seed)

    ##################################
    # 提前计算 pre_calculated_dataset ##
    ##################################
    try:
        #########################################################
        # 实验室服务器和超算上都支持直接导入 预先计算好 reference 的数据集 #
        #########################################################
        pre_calculated_dataset = datasets.DatasetDict.load_from_disk(training_args.pre_compute_reference_dataset)
        # reference_chosen_logits = [p[0] for p in pre_calculated_dataset['train'].select(range(100))['reference_chosen_logits']]
        # reference_rejected_logits = [p[0] for p in pre_calculated_dataset['train'].select(range(100))['reference_rejected_logits']]
        # import itertools
        # len1 = [[len(p1['values']) for p1 in p] for p in reference_chosen_logits]
        # len1 = list(itertools.chain(*len1))
        #
        # for l in range(1, 11):
        #     print(l, len([i for i in len1 if i == l]) / len(len1))
        #
        # len2 = [[len(p1['values']) for p1 in p] for p in reference_chosen_logits]
        # len2 = list(itertools.chain(*len2))
        # for l in range(1, 11):
        #     print(l, len([i for i in len2 if i == l]) / len(len2))
        data_args.truncation_side = "left"  # Truncate from left to ensure we don't lose labels in final turn
        tokenizer = get_tokenizer(model_args, data_args)

    except (ValueError, FileNotFoundError, TemplateError):
        ##############################
        # 超算上不支持预先计算 reference #
        assert not training_args.compute_on_8a100, "超算上不支持预先计算 reference"
        ##############################

        ####################################
        # 在实验室服务器上可以预先计算 reference #
        ####################################
        raw_datasets = get_datasets(
            data_args, shuffle=False,
            splits=data_args.dataset_splits,
            configs=data_args.dataset_configs,
            columns_to_keep=["messages", "chosen", "rejected", "prompt", "completion", "label"],
        )
        logger.info(
            f"Training on the following splits: {[split + ' : ' + str(dset.num_rows) for split, dset in raw_datasets.items()]}"
        )
        column_names = list(raw_datasets["train"].features)
        data_args.truncation_side = "left"  # Truncate from left to ensure we don't lose labels in final turn
        tokenizer = get_tokenizer(model_args, data_args)
        apply_chat_template(raw_datasets['train'][0], tokenizer=tokenizer, task="dpo",
                            auto_insert_empty_system_msg=data_args.auto_insert_empty_system_msg)
        raw_datasets = raw_datasets.map(
            apply_chat_template,
            fn_kwargs={
                "tokenizer": tokenizer,
                "task": "dpo",
                "auto_insert_empty_system_msg": data_args.auto_insert_empty_system_msg,
            },
            num_proc=1,  # data_args.preprocessing_num_workers,
            remove_columns=column_names,
            desc="Formatting comparisons with prompt template",
        )
        num_raw_train_samples = len(raw_datasets["train"])
        raw_datasets = raw_datasets.filter(
            # decontaminate_humaneval,
            fn_kwargs={"text_column": "text_chosen"},
            batched=True,
            batch_size=10_000,
            num_proc=1,
            desc="Decontaminating HumanEval samples",
        )
        num_filtered_train_samples = num_raw_train_samples - len(raw_datasets["train"])
        logger.info(
            f"Decontaminated {num_filtered_train_samples} ({num_filtered_train_samples / num_raw_train_samples * 100:.2f}%) samples from the training set."
        )

        # Replace column names with what TRL needs, text_chosen -> chosen and text_rejected -> rejected
        for split in ["train", "test"]:
            raw_datasets[split] = raw_datasets[split].rename_columns(
                {"text_prompt": "prompt", "text_chosen": "chosen", "text_rejected": "rejected"}
            )
        torch_dtype = (
            model_args.torch_dtype if model_args.torch_dtype in ["auto", None] else getattr(torch,
                                                                                            model_args.torch_dtype)
        )
        quantization_config = get_quantization_config(model_args)
        model_kwargs = dict(
            revision=model_args.model_revision,
            trust_remote_code=model_args.trust_remote_code,
            # use_flash_attention_2=model_args.use_flash_attention_2,
            torch_dtype=torch_dtype,
            use_cache=False if training_args.gradient_checkpointing else True,
            device_map=device_map,
            quantization_config=quantization_config,
        )
        model = model_args.model_name_or_path
        if is_adapter_model(model, model_args.model_revision) is True:
            logger.info(f"Loading SFT adapter for {model_args.model_name_or_path=}")
            peft_config = PeftConfig.from_pretrained(model_args.model_name_or_path, revision=model_args.model_revision)
            model_kwargs = dict(
                revision=model_args.base_model_revision,
                trust_remote_code=model_args.trust_remote_code,
                # use_flash_attention_2=model_args.use_flash_attention_2,
                torch_dtype=torch_dtype,
                use_cache=False if training_args.gradient_checkpointing else True,
                device_map=device_map,
                quantization_config=quantization_config,
            )
            base_model = AutoModelForCausalLM.from_pretrained(
                peft_config.base_model_name_or_path,
                **model_kwargs,
            )
            model = PeftModel.from_pretrained(
                base_model,
                model_args.model_name_or_path,
                revision=model_args.model_revision,
            )
            model_kwargs = None

        process_different_content = lambda row: {
            'prompt': row['prompt'] + training_args.assistant_start,
            'chosen': process_response(row['chosen'], training_args.assistant_start, training_args.assistant_end if training_args.pre_compute_reference_cut_end else None),
            'rejected': process_response(row['rejected'], training_args.assistant_start, training_args.assistant_end if training_args.pre_compute_reference_cut_end else None),
        }
        raw_datasets["train"] = raw_datasets['train'].map(process_different_content)
        raw_datasets["test"] = raw_datasets["test"].map(process_different_content)

        if training_args.debug_mode:
            raw_datasets["train"] = raw_datasets["train"].select(range(100))
            raw_datasets["test"] = raw_datasets["test"].select(range(100))

        # 在进行预先计算时，reference model设置为None，将ref model进行计算
        trainer = dpo_trainer_type(
            pre_compute_period=True,
            model=model,
            ref_model=training_args.ref_model,
            model_init_kwargs=model_kwargs,
            ref_model_init_kwargs=None,
            args=training_args,
            beta=training_args.beta,
            train_dataset=raw_datasets["train"],
            eval_dataset=raw_datasets["test"],
            tokenizer=tokenizer,
            max_length=training_args.max_length,
            max_prompt_length=training_args.max_prompt_length,
            precompute_ref_log_probs=False,
            peft_config=get_peft_config(model_args),
            loss_type=training_args.loss_type,
        )
        pre_calculated_dataset = DatasetDict({
            "train": trainer.pre_compute_and_save_reference(trainer.train_dataset,
                                                            training_args.pre_compute_reference_columns),
            "test": trainer.pre_compute_and_save_reference(trainer.eval_dataset,
                                                           training_args.pre_compute_reference_columns)})
        pre_calculated_dataset.save_to_disk(training_args.pre_compute_reference_dataset)
        del trainer, model  # 预先计算reference结束，释放空间；
        exit(0)

    #################################################
    # 在导入 or 获取了预先计算的数据集后，开始进行正式的dpo训练
    #################################################

    if training_args.pre_compute_reference_cut_end and not training_args.fast_load:
        # 重新加载text字段
        raw_datasets = get_datasets(
            data_args, shuffle=False,
            splits=data_args.dataset_splits,
            configs=data_args.dataset_configs,
            columns_to_keep=["messages", "chosen", "rejected", "prompt", "completion", "label"],
        )
        if training_args.debug_mode:
            raw_datasets["train"] = raw_datasets["train"].select(range(100))
            raw_datasets["test"] = raw_datasets["test"].select(range(100))
            pre_calculated_dataset['train'] = pre_calculated_dataset['train'].select(range(100))
            pre_calculated_dataset['test'] = pre_calculated_dataset['test'].select(range(100))

        column_names = list(raw_datasets["train"].features)
        data_args.truncation_side = "left"  # Truncate from left to ensure we don't lose labels in final turn
        tokenizer = get_tokenizer(model_args, data_args)
        apply_chat_template(raw_datasets['train'][0], tokenizer=tokenizer, task="dpo",
                            auto_insert_empty_system_msg=data_args.auto_insert_empty_system_msg)
        raw_datasets = raw_datasets.map(
            apply_chat_template,
            fn_kwargs={
                "tokenizer": tokenizer,
                "task": "dpo",
                "auto_insert_empty_system_msg": data_args.auto_insert_empty_system_msg,
            },
            num_proc=1,  # data_args.preprocessing_num_workers,
            remove_columns=column_names,
            desc="Formatting comparisons with prompt template",
        )
        raw_datasets = raw_datasets.filter(
            # decontaminate_humaneval,
            fn_kwargs={"text_column": "text_chosen"},
            batched=True,
            batch_size=10_000,
            num_proc=1,
            desc="Decontaminating HumanEval samples",
        )
        for split in ["train", "test"]:
            raw_datasets[split] = raw_datasets[split].rename_columns(
                {"text_prompt": "prompt", "text_chosen": "chosen", "text_rejected": "rejected"})
        process_different_content = lambda row: {
            'prompt': row['prompt'] + training_args.assistant_start,
            'chosen': process_response(row['chosen'], assistant_start=training_args.assistant_start),
            'rejected': process_response(row['rejected'], assistant_start=training_args.assistant_start)}
        raw_datasets["train"] = raw_datasets['train'].map(process_different_content)
        raw_datasets["test"] = raw_datasets["test"].map(process_different_content)

        pre_calculated_dataset = DatasetDict({
            "train": Dataset.from_dict({
                "prompt": raw_datasets['train']["prompt"],
                "chosen": raw_datasets['train']["chosen"],
                "rejected": raw_datasets['train']["rejected"],
                "reference_chosen_logps": pre_calculated_dataset['train']["reference_chosen_logps"],
                "reference_rejected_logps": pre_calculated_dataset['train']["reference_rejected_logps"],
                "reference_chosen_logps_per_token": pre_calculated_dataset['train']["reference_chosen_logps_per_token"],
                "reference_rejected_logps_per_token": pre_calculated_dataset['train'][
                    "reference_rejected_logps_per_token"],
                "reference_chosen_logits": pre_calculated_dataset['train']["reference_chosen_logits"],
                "reference_rejected_logits": pre_calculated_dataset['train']["reference_rejected_logits"],
                "reference_chosen_probs": pre_calculated_dataset['train']["reference_chosen_probs"],
                "reference_rejected_probs": pre_calculated_dataset['train']["reference_rejected_probs"],
                "reference_chosen_length": pre_calculated_dataset['train']["reference_chosen_length"],
                "reference_rejected_length": pre_calculated_dataset['train']["reference_rejected_length"],
            }),
            "test": Dataset.from_dict({
                "prompt": raw_datasets['test']["prompt"],
                "chosen": raw_datasets['test']["chosen"],
                "rejected": raw_datasets['test']["rejected"],
                "reference_chosen_logps": pre_calculated_dataset['test']["reference_chosen_logps"],
                "reference_rejected_logps": pre_calculated_dataset['test']["reference_rejected_logps"],
                "reference_chosen_logps_per_token": pre_calculated_dataset['test']["reference_chosen_logps_per_token"],
                "reference_rejected_logps_per_token": pre_calculated_dataset['test'][
                    "reference_rejected_logps_per_token"],
                "reference_chosen_logits": pre_calculated_dataset['test']["reference_chosen_logits"],
                "reference_rejected_logits": pre_calculated_dataset['test']["reference_rejected_logits"],
                "reference_chosen_probs": pre_calculated_dataset['test']["reference_chosen_probs"],
                "reference_rejected_probs": pre_calculated_dataset['test']["reference_rejected_probs"],
                "reference_chosen_length": pre_calculated_dataset['test']["reference_chosen_length"],
                "reference_rejected_length": pre_calculated_dataset['test']["reference_rejected_length"],
            })
        })

    # Log a few random samples from the training set:
    for index in random.sample(range(len(pre_calculated_dataset["train"])), 3):
        logger.info(
            f"Prompt sample {index} of the raw training set:\n\n{pre_calculated_dataset['train'][index]['prompt']}")
        logger.info(
            f"Chosen sample {index} of the raw training set:\n\n{pre_calculated_dataset['train'][index]['chosen']}")
        logger.info(
            f"Rejected sample {index} of the raw training set:\n\n{pre_calculated_dataset['train'][index]['rejected']}")

    torch_dtype = (
        model_args.torch_dtype if model_args.torch_dtype in ["auto", None] else getattr(torch, model_args.torch_dtype)
    )
    quantization_config = get_quantization_config(model_args)

    model_kwargs = dict(
        revision=model_args.model_revision,
        trust_remote_code=model_args.trust_remote_code,
        # use_flash_attention_2=model_args.use_flash_attention_2,
        torch_dtype=torch_dtype,
        use_cache=False if training_args.gradient_checkpointing else True,
        device_map=device_map,
        quantization_config=quantization_config,
    )
    model = model_args.model_name_or_path
    if is_adapter_model(model, model_args.model_revision) is True:
        logger.info(f"Loading SFT adapter for {model_args.model_name_or_path=}")
        peft_config = PeftConfig.from_pretrained(model_args.model_name_or_path, revision=model_args.model_revision)
        model_kwargs = dict(
            revision=model_args.base_model_revision,
            trust_remote_code=model_args.trust_remote_code,
            # use_flash_attention_2=model_args.use_flash_attention_2,
            torch_dtype=torch_dtype,
            use_cache=False if training_args.gradient_checkpointing else True,
            device_map=device_map,
            quantization_config=quantization_config,
        )
        base_model = AutoModelForCausalLM.from_pretrained(
            peft_config.base_model_name_or_path,
            **model_kwargs,
        )
        model = PeftModel.from_pretrained(
            base_model,
            model_args.model_name_or_path,
            revision=model_args.model_revision,
        )
        model_kwargs = None

    ref_model = None
    ref_model_kwargs = None

    if model_args.use_peft is True:
        ref_model = None
        ref_model_kwargs = None

    #########################
    # Instantiate DPO trainer
    #########################
    if training_args.debug_mode:
        pre_calculated_dataset["train"] = pre_calculated_dataset["train"].select(range(100))
        pre_calculated_dataset["test"] = pre_calculated_dataset["test"].select(range(100))

    trainer = dpo_trainer_type(
        model=model,
        ref_model=ref_model,
        model_init_kwargs=model_kwargs,
        ref_model_init_kwargs=ref_model_kwargs,
        args=training_args,
        beta=training_args.beta,
        train_dataset=pre_calculated_dataset["train"],
        eval_dataset=pre_calculated_dataset["test"],
        tokenizer=tokenizer,
        max_length=training_args.max_length,
        max_prompt_length=training_args.max_prompt_length,
        precompute_ref_log_probs=False,
        peft_config=get_peft_config(model_args),
        loss_type=training_args.loss_type,
        label_smoothing=training_args.label_smoothing
    )
    ###############
    # Training loop
    ###############
    if training_args.num_train_epochs > 0 or training_args.max_steps > 0:
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif last_checkpoint is not None:
            checkpoint = last_checkpoint

        if training_args.evaluation_first:
            logger.info("*** Evaluate at the beginning ***")
            metrics = trainer.evaluate()
            metrics["eval_samples"] = len(pre_calculated_dataset["test"])
            trainer.log_metrics("eval_beg", metrics)
            trainer.save_metrics("eval_beg", metrics)

        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        metrics = train_result.metrics
        metrics["train_samples"] = len(pre_calculated_dataset["train"])
        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

        logger.info("*** Training complete ***")

        ##################################
        # Save model and create model card
        ##################################
        logger.info("*** Save model ***")
        trainer.save_model(training_args.output_dir)
        # trainer.model.save_pretrained(training_args.output_dir, is_main_process=True)
        logger.info(f"Model saved to {training_args.output_dir}")

        # Save everything else on main process
        kwargs = {
            "finetuned_from": model_args.model_name_or_path,
            "dataset": list(data_args.dataset_mixer.keys()),
            "dataset_tags": list(data_args.dataset_mixer.keys()),
            "tags": ["alignment-handbook"],
        }
        if trainer.accelerator.is_main_process:
            trainer.create_model_card(**kwargs)
            # Restore k,v cache for fast inference
            trainer.model.config.use_cache = True
            trainer.model.config.save_pretrained(training_args.output_dir)

    ##########
    # Evaluate
    ##########
    if training_args.do_eval:
        logger.info("*** Evaluate ***")
        metrics = trainer.evaluate()
        metrics["eval_samples"] = len(pre_calculated_dataset["test"])
        trainer.log_metrics("eval_end", metrics)
        trainer.save_metrics("eval_end", metrics)

    if training_args.push_to_hub is True:
        logger.info("Pushing to hub...")
        trainer.push_to_hub(**kwargs)

    logger.info("*** Training complete! ***")


def run_custom_dpo_main(config_type, dpo_trainer_type):
    ###############################
    # previous setting
    ###############################
    parser = H4ArgumentParser((ModelArguments, DataArguments, config_type))
    model_args, data_args, training_args = parser.parse()

    # 警告信息，超算上不计算这些
    if training_args.compute_on_8a100:
        if training_args.pre_compute_reference_columns is not None:
            logger.warning("在超算上进行计算时，pre_compute_reference_columns 失效")
        if training_args.use_model_parallel:
            logger.warning("在超算上进行计算时，不再使用模型并行")

    # 是否进行模型并行
    device_map = None
    if get_quantization_config(model_args) is not None:
        device_map = get_kbit_device_map()
    elif not training_args.compute_on_8a100 and training_args.use_model_parallel:
        device_map = "balanced_low_0"  # "auto"

    # 是否加载 提前计算了reference信息的数据集
    if training_args.pre_compute_reference_dataset is not None:
        pre_calculate_ref_training(dpo_trainer_type, model_args, data_args, training_args, device_map)
    else:
        direct_training(dpo_trainer_type, model_args, data_args, training_args, device_map)


if __name__ == "__main__":
    run_custom_dpo_main(CustomDPOConfig, PreCalRefDpoTrainer)
